# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import io
from typing import Mapping, List, Optional, Type
import FreeCAD
import Path
from ...assets import Asset, AssetUri, AssetSerializer
from ...toolbit import ToolBit
from ...toolbit.mixins import RotaryToolBitMixin
from ..models.library import Library


class LinuxCNCSerializer(AssetSerializer):
    for_class: Type[Asset] = Library
    extensions: tuple[str] = (".tbl",)
    mime_type: str = "text/plain"
    can_import = False

    @classmethod
    def get_label(cls) -> str:
        return FreeCAD.Qt.translate("CAM", "LinuxCNC Tool Table")

    @classmethod
    def extract_dependencies(cls, data: bytes) -> List[AssetUri]:
        return []

    @classmethod
    def serialize(cls, asset: Asset) -> bytes:
        if not isinstance(asset, Library):
            raise TypeError("Asset must be a Library instance")

        output = io.BytesIO()
        for bit_no, bit in sorted(asset._bit_nos.items()):
            # Connor: assert isinstance(bit, ToolBit)
            # if not isinstance(bit, RotaryToolBitMixin):
            #     Path.Log.warning(
            #         f"Skipping too {bit.label} (bit.id) because it is not a rotary tool"
            #     )
            #     continue
            # Commenting this out. Why did we skip because it is not a rotary tool?
            diameter = bit.get_diameter().getUserPreferred()[0]
            pocket = "P0"  # TODO: is there a better way?
            # TODO: Strip units by splitting at the first space if diameter is a string
            # This is where we need a machine definition so we can export these out correctly
            # for a metric or imperial machine
            # Using user preferred for now
            if hasattr(diameter, "Value"):
                diameter_value = str(diameter.Value).replace(",", ".")
            elif isinstance(diameter, str):
                diameter_value = diameter.split(" ")[0].replace(",", ".")
            else:
                diameter_value = str(diameter).replace(",", ".")
            line = (
                f"T{bit_no} {pocket} X0 Y0 Z0 A0 B0 C0 U0 V0 W0 "
                f"D{diameter_value} I0 J0 Q0 ;{bit.label}\n"
            )
            output.write(line.encode("utf-8"))

        return output.getvalue()

    @classmethod
    def deserialize(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
    ) -> Library:
        # LinuxCNC .tbl files do not contain enough information to fully
        # reconstruct a Library and its ToolBits.
        # Therefore, deserialization is not supported.
        raise NotImplementedError("Deserialization is not supported for LinuxCNC .tbl files.")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create and edit annotation styles."""
## @package gui_annotationstyleeditor
# \ingroup draftguitools
# \brief Provides GUI tools to create and edit annotation styles.

## \addtogroup draftguitools
# @{
import json
import PySide.QtGui as QtGui
import PySide.QtWidgets as QtWidgets

from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
from draftguitools import gui_base
from draftutils import params
from draftutils import utils
from draftutils.translate import translate


class AnnotationStyleEditor(gui_base.GuiCommandSimplest):
    """Annotation style editor for text and dimensions.

    It inherits `GuiCommandSimplest` to set up the document,
    `IsActive`, and other behavior. See this class for more information.

    Attributes
    ----------
    doc: App::Document
        The active document when the command is used, so that the styles
        are saved to this document.

    styles: dict
        A dictionary with key-value pairs that define the new style.

    renamed: dict
        A dictionary that holds the name of the style that is renamed
        by the editor.

    form: PySide.QtWidgets.QDialog
        Holds the loaded interface from the `.ui` file.
    """

    def __init__(self):
        super(AnnotationStyleEditor, self).__init__(
            name=translate("draft", "Annotation Style Editor")
        )
        self.doc = None
        self.styles = {}
        self.renamed = {}
        self.current_style = None
        self.form = None

    def GetResources(self):
        """Set icon, menu and tooltip."""
        return {
            "Pixmap": ":icons/Draft_Annotation_Style.svg",
            "MenuText": QT_TRANSLATE_NOOP("Draft_AnnotationStyleEditor", "Annotation Styles"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_AnnotationStyleEditor",
                "Opens an editor to manage or create annotation styles",
            ),
        }

    def Activated(self):
        """Execute when the command is called.

        The document attribute is set here by the parent class.
        """
        super(AnnotationStyleEditor, self).Activated()
        # reset rename table and current style
        self.renamed = {}
        self.current_style = None

        # load dialog
        ui_file = ":/ui/dialog_AnnotationStyleEditor.ui"
        self.form = Gui.PySideUic.loadUi(ui_file)

        # restore stored size
        w = params.get_param("AnnotationStyleEditorWidth")
        h = params.get_param("AnnotationStyleEditorHeight")
        self.form.resize(w, h)

        # center the dialog over FreeCAD window
        mw = Gui.getMainWindow()
        self.form.move(
            mw.frameGeometry().topLeft() + mw.rect().center() - self.form.rect().center()
        )

        # set icons
        self.form.setWindowIcon(QtGui.QIcon(":/icons/Draft_Annotation_Style.svg"))
        self.form.pushButtonDelete.setIcon(QtGui.QIcon(":/icons/edit_Cancel.svg"))
        self.form.pushButtonRename.setIcon(QtGui.QIcon(":/icons/accessories-text-editor.svg"))
        self.form.pushButtonDelete.resize(self.form.pushButtonDelete.sizeHint())
        self.form.pushButtonRename.resize(self.form.pushButtonRename.sizeHint())
        self.form.pushButtonImport.setIcon(QtGui.QIcon(":/icons/Std_Import.svg"))
        self.form.pushButtonExport.setIcon(QtGui.QIcon(":/icons/Std_Export.svg"))

        # fill the styles combo
        self.styles = self.read_meta()
        for style in self.styles.keys():
            self.form.comboBoxStyles.addItem(style)

        # connect signals/slots
        self.form.comboBoxStyles.currentIndexChanged.connect(self.on_style_changed)
        self.form.pushButtonDelete.clicked.connect(self.on_delete)
        self.form.pushButtonRename.clicked.connect(self.on_rename)
        self.form.pushButtonImport.clicked.connect(self.on_import)
        self.form.pushButtonExport.clicked.connect(self.on_export)
        self.form.buttonBox.accepted.connect(self.update_style)

        # show editor dialog
        self.fill_editor()
        result = self.form.exec_()

        # process if OK was clicked
        if result:
            self.save_meta(self.styles)

        # store dialog size
        params.set_param("AnnotationStyleEditorWidth", self.form.width())
        params.set_param("AnnotationStyleEditorHeight", self.form.height())

    def read_meta(self):
        """Read the document Meta attribute and return a dict."""
        styles = {}
        meta = self.doc.Meta
        for key, value in meta.items():
            if key.startswith("Draft_Style_"):
                styles[key[12:]] = utils.repair_annotation_style(json.loads(value))
        return styles

    def save_meta(self, styles):
        """Save a dict to the document Meta attribute and update objects."""
        # save meta
        changedstyles = []
        meta = self.doc.Meta
        for key, value in styles.items():
            try:
                strvalue = json.dumps(value)
            except Exception:
                print("debug: unable to serialize this:", value)
            if "Draft_Style_" + key in meta and meta["Draft_Style_" + key] != strvalue:
                changedstyles.append(key)
            meta["Draft_Style_" + key] = strvalue

        # remove deleted styles
        todelete = []
        for key, value in meta.items():
            if key.startswith("Draft_Style_"):
                if key[12:] not in styles:
                    todelete.append(key)
        for key in todelete:
            del meta[key]

        self.doc.Meta = meta

        # propagate changes to all annotations
        for obj in self.get_annotations():
            vobj = obj.ViewObject
            try:
                current = vobj.AnnotationStyle
            except AssertionError:
                # empty annotation styles list
                pass
            else:
                if vobj.AnnotationStyle in self.renamed:
                    # the style has been renamed
                    # temporarily add the new style and switch to it
                    vobj.AnnotationStyle = [
                        vobj.AnnotationStyle,
                        self.renamed[vobj.AnnotationStyle],
                    ]
                    vobj.AnnotationStyle = self.renamed[vobj.AnnotationStyle]
                if vobj.AnnotationStyle in styles:
                    if vobj.AnnotationStyle in changedstyles:
                        # the style has changed
                        for attr, value in styles[vobj.AnnotationStyle].items():
                            if hasattr(vobj, attr):
                                try:
                                    if vobj.getTypeIdOfProperty(attr) == "App::PropertyColor":
                                        value = value | 0x000000FF
                                    setattr(vobj, attr, value)
                                except TypeError:
                                    pass
                else:
                    # the style has been removed
                    vobj.AnnotationStyle = ""
            vobj.AnnotationStyle = [""] + list(styles)

    def on_style_changed(self, index):
        """Execute as a callback when the styles combobox changes."""
        if index == 0:
            # nothing happens
            self.form.pushButtonDelete.setEnabled(False)
            self.form.pushButtonRename.setEnabled(False)
        elif index == 1:
            # Add new... entry
            reply = QtWidgets.QInputDialog.getText(
                None, translate("draft", "New Style"), translate("draft", "Style name")
            )
            if reply[1]:
                # OK or Enter pressed
                name = reply[0].strip()
                if name == "":
                    QtWidgets.QMessageBox.information(
                        None,
                        translate("draft", "Style name required"),
                        translate("draft", "No style name specified"),
                    )
                    self.form.comboBoxStyles.setCurrentIndex(0)
                elif name in self.styles:
                    QtWidgets.QMessageBox.information(
                        None,
                        translate("draft", "Style exists"),
                        translate("draft", "This style name already exists"),
                    )
                    self.form.comboBoxStyles.setCurrentIndex(0)
                else:
                    # create new style from current editor values
                    self.styles[name] = self.get_editor_values()
                    self.form.comboBoxStyles.addItem(name)
                    self.form.comboBoxStyles.setCurrentIndex(self.form.comboBoxStyles.count() - 1)
                    self.current_style = name
            else:
                # Cancel or Escape pressed
                self.form.comboBoxStyles.setCurrentIndex(0)
        else:
            # Existing style
            if self.current_style is not None:
                # save editor values to current style first
                self.styles[self.current_style] = self.get_editor_values()
            self.current_style = self.form.comboBoxStyles.itemText(index)
            self.fill_editor(self.current_style)
            self.form.pushButtonDelete.setEnabled(True)
            self.form.pushButtonRename.setEnabled(True)

    def on_delete(self):
        """Execute as a callback when the delete button is pressed."""
        index = self.form.comboBoxStyles.currentIndex()
        style = self.form.comboBoxStyles.itemText(index)

        if self.get_style_users(style):
            reply = QtWidgets.QMessageBox.question(
                None,
                translate("draft", "Style in use"),
                translate("draft", "This style is used by some objects in this document. Proceed?"),
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.No,
            )
            if reply == QtWidgets.QMessageBox.No:
                return

        del self.styles[style]
        # We need to reset self.current_style, which is the deleted style,
        # to stop on_style_changed from adding that style again:
        self.current_style = None

        self.form.comboBoxStyles.currentIndexChanged.disconnect(self.on_style_changed)
        self.form.comboBoxStyles.removeItem(index)
        if not self.styles:
            self.form.comboBoxStyles.setCurrentIndex(0)
        # Update the dialog and self.current_style:
        self.on_style_changed(self.form.comboBoxStyles.currentIndex())
        self.form.comboBoxStyles.currentIndexChanged.connect(self.on_style_changed)

    def on_rename(self):
        """Execute as a callback when the rename button is pressed."""
        index = self.form.comboBoxStyles.currentIndex()
        style = self.form.comboBoxStyles.itemText(index)

        reply = QtWidgets.QInputDialog.getText(
            None,
            translate("draft", "Rename Style"),
            translate("draft", "New name"),
            QtWidgets.QLineEdit.Normal,
            style,
        )
        if reply[1]:
            # OK or Enter pressed
            newname = reply[0]
            if newname in self.styles:
                reply = QtWidgets.QMessageBox.information(
                    None,
                    translate("draft", "Style exists"),
                    translate("draft", "This style name already exists"),
                )
            else:
                self.form.comboBoxStyles.setItemText(index, newname)
                value = self.styles[style]
                del self.styles[style]
                self.styles[newname] = value
                self.renamed[style] = newname
                self.current_style = newname

    def on_import(self):
        """Import styles from a json file."""
        filename = QtWidgets.QFileDialog.getOpenFileName(
            QtWidgets.QApplication.activeWindow(),
            translate("draft", "Open Styles File"),
            None,
            translate("draft", "JSON files (*.json *.JSON)"),
        )
        if filename and filename[0]:
            nstyles = {}
            with open(filename[0]) as f:
                for key, val in json.load(f).items():
                    nstyles[key] = utils.repair_annotation_style(val)
            if nstyles:
                self.styles.update(nstyles)
                for style in self.styles.keys():
                    if self.form.comboBoxStyles.findText(style) == -1:
                        self.form.comboBoxStyles.addItem(style)
                self.fill_editor(self.current_style)  # The current style may have changed.
                print("Styles updated from " + filename[0])

    def on_export(self):
        """Export styles to a json file."""
        filename = QtWidgets.QFileDialog.getSaveFileName(
            QtWidgets.QApplication.activeWindow(),
            translate("draft", "Save Styles File"),
            None,
            translate("draft", "JSON file (*.json)"),
        )
        if filename and filename[0]:
            self.update_style()
            with open(filename[0], "w") as f:
                json.dump(self.styles, f, indent=4)
            print("Styles saved to " + filename[0])

    def fill_editor(self, style=None):
        """Fill the editor fields with the contents of a style."""
        default = utils.get_default_annotation_style()
        if style is None or style == "":
            style = {}
            for key, val in default.items():
                style[key] = val[1]
        elif isinstance(style, dict):
            pass
        elif isinstance(style, str) and style in self.styles:
            style = self.styles[style]
        else:
            print("debug: unable to fill dialog from style", style)

        for key, value in style.items():
            control = getattr(self.form, key)
            if default[key][0] == "str":
                control.setText(value)
            elif default[key][0] == "font":
                control.setCurrentFont(QtGui.QFont(value))
            elif default[key][0] == "color":
                color = QtGui.QColor(utils.rgba_to_argb(value))
                control.setProperty("color", color)
            elif default[key][0] == "int":
                control.setValue(value)
            elif default[key][0] == "float":
                if hasattr(control, "setText"):
                    control.setText(App.Units.Quantity(value, App.Units.Length).UserString)
                else:
                    control.setValue(value)
            elif default[key][0] == "bool":
                control.setChecked(value)
            elif default[key][0] == "index":
                control.setCurrentIndex(value)

    def update_style(self):
        """Update the current style with the values from the editor."""
        index = self.form.comboBoxStyles.currentIndex()
        if index > 1:
            style = self.form.comboBoxStyles.itemText(index)
            self.styles[style] = self.get_editor_values()

    def get_editor_values(self):
        default = utils.get_default_annotation_style()
        values = {}
        for key in default.keys():
            control = getattr(self.form, key)
            if default[key][0] == "str":
                values[key] = control.text()
            elif default[key][0] == "font":
                values[key] = control.currentFont().family()
            elif default[key][0] == "color":
                values[key] = utils.argb_to_rgba(control.property("color").rgba())
            elif default[key][0] == "int":
                values[key] = control.value()
            elif default[key][0] == "float":
                if hasattr(control, "setText"):
                    values[key] = App.Units.Quantity(control.text()).Value
                else:
                    values[key] = control.value()
            elif default[key][0] == "bool":
                values[key] = control.isChecked()
            elif default[key][0] == "index":
                values[key] = control.currentIndex()
        return values

    def get_annotations(self):
        """Get all objects that support annotation styles."""
        users = []
        for obj in self.doc.Objects:
            vobj = obj.ViewObject
            if "AnnotationStyle" in vobj.PropertiesList:
                users.append(obj)
        return users

    def get_style_users(self, style):
        """Get all objects using a certain style."""
        users = []
        for obj in self.get_annotations():
            if obj.ViewObject.AnnotationStyle == style:
                users.append(obj)
        return users


Gui.addCommand("Draft_AnnotationStyleEditor", AnnotationStyleEditor())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Furgo                                              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import unittest
import FreeCAD
import FreeCADGui
from bimtests.TestArchBase import TestArchBase


class TestArchBaseGui(TestArchBase):
    """
    The base class for all Arch/BIM GUI unit tests.
    It inherits from TestArchBase to handle document setup and adds
    GUI-specific initialization by activating the BIM workbench.
    """

    @classmethod
    def setUpClass(cls):
        """
        Ensure the GUI is available and activate the BIM workbench once
        before any tests in the inheriting class are run.
        """
        if not FreeCAD.GuiUp:
            raise unittest.SkipTest("Cannot run GUI tests in a CLI environment.")

        # Activating the workbench ensures all GUI commands are loaded and ready.
        # TODO: commenting out this line for now as it causes a timeout without further logging in
        # CI
        # FreeCADGui.activateWorkbench("BIMWorkbench")

    def setUp(self):
        """
        Run the parent's setup to create the uniquely named document.
        The workbench is already activated by setUpClass.
        """
        super().setUp()

    def pump_gui_events(self, timeout_ms=200):
        """Run the Qt event loop briefly so queued GUI callbacks execute.

        This helper starts a QEventLoop and quits it after `timeout_ms` milliseconds using
        QTimer.singleShot. Any exception (e.g. missing Qt in the environment) is silently ignored so
        tests can still run in pure-CLI environments where the GUI isn't available.
        """
        try:
            from PySide import QtCore

            loop = QtCore.QEventLoop()
            QtCore.QTimer.singleShot(int(timeout_ms), loop.quit)
            loop.exec_()
        except Exception:
            # Best-effort: if Qt isn't present or event pumping fails, continue.
            pass



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM workbench"""

import os

import FreeCAD
import FreeCADGui
import Arch_rc


class BIMWorkbench(Workbench):

    def __init__(self):

        def QT_TRANSLATE_NOOP(context, text):
            return text

        bdir = os.path.join(FreeCAD.getResourceDir(), "Mod", "BIM")
        tt = QT_TRANSLATE_NOOP("BIM", "The BIM workbench is used to model buildings")
        self.__class__.MenuText = QT_TRANSLATE_NOOP("BIM", "BIM")
        self.__class__.ToolTip = tt
        self.__class__.Icon = os.path.join(bdir, "Resources", "icons", "BIMWorkbench.svg")

    def Initialize(self):

        # add translations and icon paths
        FreeCADGui.addIconPath(":/icons")
        FreeCADGui.addLanguagePath(":/translations")

        # Create menus and toolbars
        self.createTools()

        # Load Arch & Draft preference pages
        self.loadPreferences()

        Log("Loading BIM module… done\n")
        FreeCADGui.updateLocale()

    def createTools(self):
        "Create tolbars and menus"

        def QT_TRANSLATE_NOOP(context, text):
            return text

        # Import Draft & BIM commands
        import DraftTools
        import bimcommands
        from nativeifc import ifc_commands

        # build menus and toolbars
        self.draftingtools = [
            "BIM_Sketch",
            "Draft_Line",
            "Draft_Wire",
            "Draft_Circle",
            "Draft_Arc",
            "Draft_Arc_3Points",
            "Draft_Fillet",
            "Draft_Ellipse",
            "Draft_Polygon",
            "Draft_Rectangle",
            "Draft_BSpline",
            "Draft_BezCurve",
            "Draft_CubicBezCurve",
            "Draft_Point",
        ]

        self.annotationtools = [
            "BIM_Text",
            "Draft_ShapeString",
            "BIM_DimensionAligned",
            "BIM_DimensionHorizontal",
            "BIM_DimensionVertical",
            "BIM_Leader",
            "Draft_Label",
            "Draft_Hatch",
            "Arch_Axis",
            "Arch_AxisSystem",
            "Arch_Grid",
            "Arch_SectionPlane",
            "BIM_TDPage",
            "BIM_TDView",
        ]

        self.bimtools = [
            "Arch_Site",
            "Arch_Building",
            "Arch_Level",
            "Arch_Space",
            "Separator",
            "Arch_Wall",
            "Arch_CurtainWall",
            "BIM_Column",
            "BIM_Beam",
            "BIM_Slab",
            "BIM_Door",
            "Arch_Window",
            "Arch_Pipe",
            "Arch_PipeConnector",
            "Arch_Stairs",
            "Arch_Roof",
            "Arch_Panel",
            "Arch_Frame",
            "Arch_Fence",
            "Arch_Truss",
            "Arch_Equipment",
            "Arch_Rebar",
        ]

        self.modify_gen = [
            "Draft_Move",
            "BIM_Copy",
            "Draft_Rotate",
            "BIM_Clone",
            "BIM_SimpleCopy",
            "BIM_Compound",
        ]
        self.modify_2d = [
            "Draft_Offset",
            "BIM_Offset2D",
            "Draft_Trimex",
            "Draft_Join",
            "Draft_Split",
            "Draft_Scale",
            "Draft_Stretch",
            "Draft_Draft2Sketch",
        ]
        self.modify_obj = [
            "Draft_Upgrade",
            "Draft_Downgrade",
            "Arch_Add",
            "Arch_Remove",
        ]
        self.modify_3d = [
            "Draft_OrthoArray",
            "Draft_PathArray",
            "Draft_PolarArray",
            "Draft_PointArray",
            "Arch_CutPlane",
            "Draft_Mirror",
            "BIM_Extrude",
            "BIM_Cut",
            "BIM_Fuse",
            "BIM_Common",
        ]

        sep = ["Separator"]
        self.modify = (
            self.modify_gen + sep + self.modify_2d + sep + self.modify_obj + sep + self.modify_3d
        )

        self.manage = [
            "BIM_Setup",
            "BIM_Views",
            "BIM_ProjectManager",
            "BIM_Windows",
            "BIM_IfcElements",
            "BIM_IfcQuantities",
            "BIM_IfcProperties",
            "BIM_Classification",
            "BIM_Layers",
            "BIM_Material",
            "Arch_Schedule",
            "BIM_Preflight",
            "Draft_AnnotationStyleEditor",
        ]

        self.utils = [
            "BIM_TogglePanels",
            "BIM_Trash",
            "BIM_WPView",
            "Draft_SelectGroup",
            "Draft_Slope",
            "Draft_WorkingPlaneProxy",
            "Draft_AddConstruction",
            "Arch_SplitMesh",
            "Arch_MeshToShape",
            "Arch_SelectNonSolidMeshes",
            "Arch_RemoveShape",
            "Arch_CloseHoles",
            "Arch_MergeWalls",
            "Arch_Check",
            "Arch_ToggleIfcBrepFlag",
            "Arch_ToggleSubs",
            "Arch_Survey",
            "BIM_Diff",
            "BIM_IfcExplorer",
            "Arch_IfcSpreadsheet",
            "BIM_ImagePlane",
            "BIM_Unclone",
            "BIM_Rewire",
            "BIM_Glue",
            "BIM_Reextrude",
            "Arch_PanelTools",
            "Arch_StructureTools",
            "BIM_Project",
        ]

        nudge = [
            "BIM_Nudge_Switch",
            "BIM_Nudge_Up",
            "BIM_Nudge_Down",
            "BIM_Nudge_Left",
            "BIM_Nudge_Right",
            "BIM_Nudge_RotateLeft",
            "BIM_Nudge_RotateRight",
            "BIM_Nudge_Extend",
            "BIM_Nudge_Shrink",
        ]

        # append BIM snaps

        from draftutils import init_tools

        self.snapbar = init_tools.get_draft_snap_commands()
        self.snapmenu = self.snapbar + [
            "BIM_SetWPFront",
            "BIM_SetWPTop",
            "BIM_SetWPSide",
            "Draft_SelectPlane",
        ]

        # create generic tools command
        class BIM_GenericTools:
            def GetCommands(self):
                return (
                    "Arch_Profile",
                    "BIM_Box",
                    "BIM_Builder",
                    "Draft_Facebinder",
                    "BIM_Library",
                    "Arch_Component",
                    "Arch_Reference",
                )

            def GetResources(self):
                t = QT_TRANSLATE_NOOP("BIM_GenericTools", "Generic 3D Tools")
                return {"MenuText": t, "ToolTip": t, "Icon": "BIM_Box"}

            def IsActive(self):
                v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
                return v

        FreeCADGui.addCommand("BIM_GenericTools", BIM_GenericTools())
        self.bimtools.append("BIM_GenericTools")

        # create create 2D views command
        class BIM_Create2DViews:
            def GetCommands(self):
                return ("BIM_DrawingView", "BIM_Shape2DView", "BIM_Shape2DCut")

            def GetResources(self):
                t = QT_TRANSLATE_NOOP("BIM_Create2DViews", "Create 2D Views")
                return {"MenuText": t, "ToolTip": t, "Icon": "BIM_DrawingView"}

            def IsActive(self):
                v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
                return v

        FreeCADGui.addCommand("BIM_Create2DViews", BIM_Create2DViews())
        insert_at_index = self.annotationtools.index("BIM_TDPage")
        self.annotationtools.insert(insert_at_index, "BIM_Create2DViews")

        # load rebar tools (Reinforcement addon)

        try:
            import RebarTools
        except ImportError:
            RebarGroupCommand = None  # for workaround for issue #26539 and #27984
        else:
            # create popup group for Rebar tools
            class RebarGroupCommand:
                def GetCommands(self):
                    return tuple(["Arch_Rebar"] + RebarTools.RebarCommands)

                def GetResources(self):
                    return {
                        "MenuText": QT_TRANSLATE_NOOP("Arch_RebarTools", "Reinforcement Tools"),
                        "ToolTip": QT_TRANSLATE_NOOP("Arch_RebarTools", "Reinforcement tools"),
                        "Icon": "Arch_Rebar",
                    }

                def IsActive(self):
                    v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
                    return v

            FreeCADGui.addCommand("Arch_RebarTools", RebarGroupCommand())
            self.bimtools[self.bimtools.index("Arch_Rebar")] = "Arch_RebarTools"
            RebarTools.load_translations()
            Log("Load Reinforcement Module… done\n")
            if hasattr(RebarTools, "updateLocale"):
                RebarTools.updateLocale()
            # self.rebar = RebarTools.RebarCommands + ["Arch_Rebar"]

        # load Reporting

        try:
            import report
        except ImportError:
            pass
        else:
            if "Report_Create" in Gui.listCommands():
                self.manage[self.manage.index("Arch_Schedule")] = "Report_Create"

        # load webtools

        try:
            import BIMServer
            import Git
            import Sketchfab
        except ImportError:
            pass
        else:
            self.utils.extend(
                [
                    "WebTools_Git",
                    "WebTools_BimServer",
                    "WebTools_Sketchfab",
                ]
            )

        # load flamingo

        try:
            import CommandsPolar
            import CommandsFrame
            import CommandsPipe
        except ImportError:
            flamingo = None
        else:
            flamingo = [
                "frameIt",
                "fillFrame",
                "insertPath",
                "insertSection",
                "FrameLineManager",
                "spinSect",
                "reverseBeam",
                "shiftBeam",
                "pivotBeam",
                "levelBeam",
                "alignEdge",
                "rotJoin",
                "alignFlange",
                "stretchBeam",
                "extend",
                "adjustFrameAngle",
                "insertPipe",
                "insertElbow",
                "insertReduct",
                "insertCap",
                "insertFlange",
                "insertUbolt",
                "insertPypeLine",
                "breakPipe",
                "mateEdges",
                "extend2intersection",
                "extend1intersection",
                "laydown",
                "raiseup",
            ]

        # load fasteners

        try:
            import FastenerBase
            import FastenersCmd
        except ImportError:
            fasteners = None
        else:
            fasteners = [
                c for c in FastenerBase.FSGetCommands("screws") if not isinstance(c, tuple)
            ]

        # load nativeifc tools

        ifctools = ifc_commands.get_commands()

        # create toolbars

        t1 = QT_TRANSLATE_NOOP("Workbench", "Drafting Tools")
        t2 = QT_TRANSLATE_NOOP("Workbench", "Draft Snap")
        t3 = QT_TRANSLATE_NOOP("Workbench", "3D/BIM Tools")
        t4 = QT_TRANSLATE_NOOP("Workbench", "Annotation Tools")
        t5 = QT_TRANSLATE_NOOP("Workbench", "2D Tools")
        t6 = QT_TRANSLATE_NOOP("Workbench", "Manage Tools")
        t7 = QT_TRANSLATE_NOOP("Workbench", "General Tools")
        t8 = QT_TRANSLATE_NOOP("Workbench", "Object Tools")
        t9 = QT_TRANSLATE_NOOP("Workbench", "3D Tools")
        self.appendToolbar(t1, self.draftingtools)
        self.appendToolbar(t2, self.snapbar)
        self.appendToolbar(t3, self.bimtools)
        self.appendToolbar(t4, self.annotationtools)
        self.appendToolbar(t7, self.modify_gen)
        self.appendToolbar(t5, self.modify_2d)
        self.appendToolbar(t8, self.modify_obj)
        self.appendToolbar(t9, self.modify_3d)
        self.appendToolbar(t6, self.manage)

        # create menus

        t1 = QT_TRANSLATE_NOOP("Workbench", "&2D Drafting")
        t2 = QT_TRANSLATE_NOOP("Workbench", "&3D/BIM")
        t3 = QT_TRANSLATE_NOOP("Workbench", "Reinforcement Tools")
        t4 = QT_TRANSLATE_NOOP("Workbench", "&Annotation")
        t5 = QT_TRANSLATE_NOOP("Workbench", "&Snapping")
        t6 = QT_TRANSLATE_NOOP("Workbench", "&Modify")
        t7 = QT_TRANSLATE_NOOP("Workbench", "&Manage")
        # t8 =  QT_TRANSLATE_NOOP("Workbench", "&IFC")
        t9 = QT_TRANSLATE_NOOP("Workbench", "&Flamingo")
        t10 = QT_TRANSLATE_NOOP("Workbench", "&Fasteners")
        t11 = QT_TRANSLATE_NOOP("Workbench", "&Utils")
        t12 = QT_TRANSLATE_NOOP("Workbench", "Nudge")

        # self.bimtools_menu = list(self.bimtools)
        # if "Arch_RebarTools" in self.bimtools_menu:
        #    self.bimtools_menu.remove("Arch_RebarTools")
        self.appendMenu(t1, self.draftingtools)
        self.appendMenu(t2, self.bimtools)
        # if self.rebar:
        #    self.appendMenu([t2, t3], self.rebar)
        self.appendMenu(t4, self.annotationtools)
        self.appendMenu(t5, self.snapmenu)
        self.appendMenu(t6, self.modify)
        self.appendMenu(t7, self.manage)
        # if ifctools:
        #    self.appendMenu(t8, ifctools)
        if flamingo:
            self.appendMenu(t9, flamingo)
        if fasteners:
            self.appendMenu(t10, fasteners)
        self.appendMenu(t11, self.utils + ifctools)
        self.appendMenu([t11, t12], nudge)

        # workaround for issue #26539 and #27984:
        # create tool lists without grouped commands for TaskWatcher
        # https://github.com/FreeCAD/FreeCAD/issues/26539
        # https://github.com/FreeCAD/FreeCAD/issues/27984
        chk = (
            ("Arch_RebarTools", RebarGroupCommand),
            ("BIM_Create2DViews", BIM_Create2DViews),
            ("BIM_GenericTools", BIM_GenericTools),
        )
        for attr in ("draftingtools", "annotationtools", "bimtools", "modify"):
            lst = getattr(self, attr)
            for itm in chk:
                if not itm[0] in lst:
                    continue
                idx = lst.index(itm[0])
                cmds = list(itm[1].GetCommands(itm[1]))
                lst = lst[:idx] + cmds + lst[idx + 1 :]
            setattr(self, attr, lst)

    def loadPreferences(self):
        """Set up preferences pages"""

        def QT_TRANSLATE_NOOP(context, text):
            return text

        t1 = QT_TRANSLATE_NOOP("QObject", "BIM")
        t2 = QT_TRANSLATE_NOOP("QObject", "Draft")
        FreeCADGui.addPreferencePage(":/ui/preferences-arch.ui", t1)
        FreeCADGui.addPreferencePage(":/ui/preferences-archdefaults.ui", t1)
        FreeCADGui.addPreferencePage(":/ui/preferencesNativeIFC.ui", t1)
        if hasattr(FreeCADGui, "draftToolBar"):
            if hasattr(FreeCADGui.draftToolBar, "loadedPreferences"):
                return
        from draftutils import params

        params._param_observer_start()
        FreeCADGui.addPreferencePage(":/ui/preferences-draft.ui", t2)
        FreeCADGui.addPreferencePage(":/ui/preferences-draftinterface.ui", t2)
        FreeCADGui.addPreferencePage(":/ui/preferences-draftsnap.ui", t2)
        FreeCADGui.addPreferencePage(":/ui/preferences-draftvisual.ui", t2)
        FreeCADGui.addPreferencePage(":/ui/preferences-drafttexts.ui", t2)
        FreeCADGui.draftToolBar.loadedPreferences = True

    def setupMultipleObjectSelection(self):

        import BimSelect

        if hasattr(FreeCADGui, "addDocumentObserver") and not hasattr(self, "BimSelectObserver"):
            self.BimSelectObserver = BimSelect.Setup()
            FreeCADGui.addDocumentObserver(self.BimSelectObserver)

    def Activated(self):

        import WorkingPlane
        from draftutils import todo
        import BimStatus
        from nativeifc import ifc_observer
        from draftutils import grid_observer

        PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")

        if hasattr(FreeCADGui, "draftToolBar"):
            FreeCADGui.draftToolBar.Activated()
        if hasattr(FreeCADGui, "Snapper"):
            FreeCADGui.Snapper.show()
        WorkingPlane._view_observer_start()
        grid_observer._view_observer_setup()

        if PARAMS.GetBool("FirstTime", True) and (not hasattr(FreeCAD, "TestEnvironment")):
            todo.ToDo.delay(FreeCADGui.runCommand, "BIM_Welcome")
        todo.ToDo.delay(BimStatus.setStatusIcons, True)
        FreeCADGui.Control.clearTaskWatcher()

        class BimWatcher:
            def __init__(self, cmds, name, invert=False):
                self.commands = cmds
                self.title = name
                self.invert = invert

            def shouldShow(self):
                if self.invert:
                    return (FreeCAD.ActiveDocument != None) and (
                        FreeCADGui.Selection.getSelection() != []
                    )
                else:
                    return (FreeCAD.ActiveDocument != None) and (
                        not FreeCADGui.Selection.getSelection()
                    )

        FreeCADGui.Control.addTaskWatcher(
            [
                BimWatcher(self.draftingtools + self.annotationtools, "2D geometry"),
                BimWatcher(self.bimtools, "3D/BIM geometry"),
                BimWatcher(self.modify, "Modify", invert=True),
            ]
        )

        # restore views widget if needed
        if PARAMS.GetBool("RestoreBimViews", True):
            from bimcommands import BimViews

            w = BimViews.findWidget()
            if not w:
                FreeCADGui.runCommand("BIM_Views")
            else:
                w.show()
                w.toggleViewAction().setVisible(True)

        self.setupMultipleObjectSelection()

        # add NativeIFC document observer
        ifc_observer.add_observer()

        # adding a Help menu manipulator
        # https://github.com/FreeCAD/FreeCAD/pull/10933
        class BIM_WBManipulator:
            def modifyMenuBar(self):
                return [
                    {"insert": "BIM_Examples", "menuItem": "Std_ReportBug", "after": ""},
                    {"insert": "BIM_Tutorial", "menuItem": "Std_ReportBug", "after": ""},
                    {"insert": "BIM_Help", "menuItem": "Std_ReportBug", "after": ""},
                    {"insert": "BIM_Welcome", "menuItem": "Std_ReportBug", "after": ""},
                ]

        reload = hasattr(Gui, "BIM_WBManipulator")  # BIM WB has previously been loaded.
        if not getattr(Gui, "BIM_WBManipulator", None):
            Gui.BIM_WBManipulator = BIM_WBManipulator()
        Gui.addWorkbenchManipulator(Gui.BIM_WBManipulator)
        if reload:
            Gui.activeWorkbench().reloadActive()

        Log("BIM workbench activated\n")

    def Deactivated(self):

        from draftutils import todo
        import BimStatus
        from bimcommands import BimViews
        import WorkingPlane
        from nativeifc import ifc_observer
        from draftutils import grid_observer

        PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")

        if hasattr(self, "BimSelectObserver"):
            FreeCADGui.removeDocumentObserver(self.BimSelectObserver)
            del self.BimSelectObserver

        if hasattr(FreeCADGui, "draftToolBar"):
            FreeCADGui.draftToolBar.Deactivated()
        if hasattr(FreeCADGui, "Snapper"):
            FreeCADGui.Snapper.hide()
        WorkingPlane._view_observer_stop()
        grid_observer._view_observer_setup()

        # print("Deactivating status icon")
        todo.ToDo.delay(BimStatus.setStatusIcons, False)
        FreeCADGui.Control.clearTaskWatcher()

        # store views widget state and vertical size
        w = BimViews.findWidget()
        if w:
            PARAMS.SetBool("RestoreBimViews", w.isVisible())
            PARAMS.SetInt("BimViewsSize", w.height())
            w.hide()
            w.toggleViewAction().setVisible(False)

        # add NativeIFC document observer
        ifc_observer.remove_observer()

        # Ifc stuff
        try:
            from nativeifc import ifc_status

            ifc_status.toggle_lock(False)
        except:
            pass

        # remove manipulator
        if hasattr(Gui, "BIM_WBManipulator"):
            Gui.removeWorkbenchManipulator(Gui.BIM_WBManipulator)
            Gui.BIM_WBManipulator = None
            Gui.activeWorkbench().reloadActive()

        Log("BIM workbench deactivated\n")

    def ContextMenu(self, recipient):

        import DraftTools

        translate = FreeCAD.Qt.translate

        if recipient == "Tree":
            groups = False
            ungroupable = False
            for o in FreeCADGui.Selection.getSelection():
                if o.isDerivedFrom("App::DocumentObjectGroup") or o.hasExtension(
                    "App::GroupExtension"
                ):
                    groups = True
                else:
                    groups = False
                    break
            for o in FreeCADGui.Selection.getSelection():
                for parent in o.InList:
                    if parent.isDerivedFrom("App::DocumentObjectGroup") or parent.hasExtension(
                        "App::GroupExtension"
                    ):
                        if o in parent.Group:
                            ungroupable = True
                        else:
                            ungroupable = False
                            break
            if groups:
                self.appendContextMenu("", ["Draft_SelectGroup"])
            if ungroupable:
                self.appendContextMenu("", ["BIM_Ungroup"])
            if (len(FreeCADGui.Selection.getSelection()) == 1) and (
                FreeCADGui.Selection.getSelection()[0].Name == "Trash"
            ):
                self.appendContextMenu("", ["BIM_EmptyTrash"])
        elif recipient == "View":
            self.appendContextMenu(translate("BIM", "Snapping"), self.snapmenu)
        if FreeCADGui.Selection.getSelection():
            if FreeCADGui.Selection.getSelection()[0].Name != "Trash":
                self.appendContextMenu("", ["BIM_Trash"])
            self.appendContextMenu("", ["Draft_AddConstruction", "Draft_AddToGroup"])
            allclones = False
            for obj in FreeCADGui.Selection.getSelection():
                if hasattr(obj, "CloneOf") and obj.CloneOf:
                    allclones = True
                else:
                    allclones = False
                    break
            if allclones:
                self.appendContextMenu("", ["BIM_ResetCloneColors"])
            if len(FreeCADGui.Selection.getSelection()) == 1:
                obj = FreeCADGui.Selection.getSelection()[0]
                if hasattr(obj, "Group"):
                    if obj.getTypeIdOfProperty("Group") == "App::PropertyLinkList":
                        self.appendContextMenu("", ["BIM_Reorder"])
                if obj.isDerivedFrom("TechDraw::DrawView"):
                    self.appendContextMenu("", ["BIM_MoveView"])

    def GetClassName(self):
        return "Gui::PythonWorkbench"


FreeCADGui.addWorkbench(BIMWorkbench)

# Preference pages for importing and exporting various file formats
# are independent of the loading of the workbench and can be loaded at startup


def QT_TRANSLATE_NOOP(context, text):
    return text


t = QT_TRANSLATE_NOOP("QObject", "Import-Export")
FreeCADGui.addPreferencePage(":/ui/preferences-ifc.ui", t)
FreeCADGui.addPreferencePage(":/ui/preferences-ifc-export.ui", t)
FreeCADGui.addPreferencePage(":/ui/preferences-dae.ui", t)
FreeCADGui.addPreferencePage(":/ui/preferences-sh3d-import.ui", t)
FreeCADGui.addPreferencePage(":/ui/preferences-webgl.ui", t)

# Add unit tests
FreeCAD.__unit_test__ += ["TestArchGui"]
# The NativeIFC tests require internet connection and file download
# FreeCAD.__unit_test__ += ["nativeifc.ifc_selftest"]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""BIM Rebar command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Rebar:
    "the Arch Rebar command definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Rebar",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Rebar", "Custom Rebar"),
            "Accel": "R, B",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Rebar",
                "Creates a reinforcement bar from the selected face of solid object and/or a sketch",
            ),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        import ArchComponent

        sel = FreeCADGui.Selection.getSelectionEx()
        if sel:
            obj = sel[0].Object
            if hasattr(obj, "Shape") and obj.Shape.Solids:
                # this is our host object
                if len(sel) > 1:
                    sk = sel[1].Object
                    if hasattr(sk, "Shape"):
                        if len(sk.Shape.Wires) == 1:
                            # we have a structure and a wire: create the rebar now
                            FreeCAD.ActiveDocument.openTransaction(
                                translate("Arch", "Create Rebar")
                            )
                            FreeCADGui.addModule("Arch")
                            FreeCADGui.doCommand(
                                "Arch.makeRebar(FreeCAD.ActiveDocument."
                                + obj.Name
                                + ",FreeCAD.ActiveDocument."
                                + sk.Name
                                + ")"
                            )
                            FreeCAD.ActiveDocument.commitTransaction()
                            FreeCAD.ActiveDocument.recompute()
                            return
                else:
                    # we have only a structure: open the sketcher
                    FreeCADGui.activateWorkbench("SketcherWorkbench")
                    FreeCADGui.runCommand("Sketcher_NewSketch")
                    FreeCAD.ArchObserver = ArchComponent.ArchSelectionObserver(
                        obj,
                        FreeCAD.ActiveDocument.Objects[-1],
                        hide=False,
                        nextCommand="Arch_Rebar",
                    )
                    FreeCADGui.Selection.addObserver(FreeCAD.ArchObserver)
                    return
            elif hasattr(obj, "Shape"):
                if len(obj.Shape.Wires) == 1:
                    # we have only a wire: extract its support object, if available, and make the rebar
                    support = "None"
                    if hasattr(obj, "AttachmentSupport"):
                        if obj.AttachmentSupport:
                            if len(obj.AttachmentSupport) != 0:
                                support = (
                                    "FreeCAD.ActiveDocument." + obj.AttachmentSupport[0][0].Name
                                )
                    FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Rebar"))
                    FreeCADGui.addModule("Arch")
                    FreeCADGui.doCommand(
                        "Arch.makeRebar(" + support + ",FreeCAD.ActiveDocument." + obj.Name + ")"
                    )
                    FreeCAD.ActiveDocument.commitTransaction()
                    FreeCAD.ActiveDocument.recompute()
                    return

        FreeCAD.Console.PrintMessage(
            translate("Arch", "Select a base face on a structural object") + "\n"
        )
        FreeCADGui.Control.closeDialog()
        FreeCADGui.Control.showDialog(ArchComponent.SelectionTaskPanel())
        FreeCAD.ArchObserver = ArchComponent.ArchSelectionObserver(nextCommand="Arch_Rebar")
        FreeCADGui.Selection.addObserver(FreeCAD.ArchObserver)


FreeCADGui.addCommand("Arch_Rebar", Arch_Rebar())



# ***************************************************************************
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint centrif task panel for the document object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package task_constraint_centrif
#  \ingroup FEM
#  \brief task panel for constraint centrif object

from PySide import QtCore
from PySide import QtGui

import FreeCAD
import FreeCADGui

from femguiutils import selection_widgets
from . import base_femtaskpanel


class _TaskPanel(base_femtaskpanel._BaseTaskPanel):
    """
    The TaskPanel for editing References property of FemConstraintCentrif objects
    """

    def __init__(self, obj):
        super().__init__(obj)

        # parameter widget
        self.parameter_widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/ConstraintCentrif.ui"
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_rotation_frequency,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.rotation_frequency_changed,
        )
        self.init_parameter_widget()

        # axis of rotation selection widget
        self.axis_selection_widget = selection_widgets.GeometryElementsSelection(
            obj.RotationAxis, ["Edge"], False, False
        )
        self.axis_selection_widget.setWindowTitle(
            self.axis_selection_widget.tr("Axis Reference Selector")
        )

        # loaded body selection widget
        self.body_selection_widget = selection_widgets.GeometryElementsSelection(
            obj.References, ["Solid", "Face"], False, False
        )

        # form made from param and selection widget
        self.form = [self.parameter_widget, self.body_selection_widget, self.axis_selection_widget]

    def accept(self):
        # check values RotationAxis
        items = len(self.axis_selection_widget.references)
        FreeCAD.Console.PrintMessage(
            "Task panel: found axis references: {}\n{}\n".format(
                items, self.axis_selection_widget.references
            )
        )

        if items != 1:
            msgBox = QtGui.QMessageBox()
            msgBox.setIcon(QtGui.QMessageBox.Question)
            msgBox.setText(
                f"Constraint Centrif requires exactly one line\n\nfound references: {items}"
            )
            msgBox.setWindowTitle("FreeCAD FEM Constraint Centrif - Axis Selection")
            retryButton = msgBox.addButton(QtGui.QMessageBox.Retry)
            ignoreButton = msgBox.addButton(QtGui.QMessageBox.Ignore)
            msgBox.exec_()

            if msgBox.clickedButton() == retryButton:
                return False
            elif msgBox.clickedButton() == ignoreButton:
                pass

        # check values BodyReference
        items = len(self.body_selection_widget.references)
        FreeCAD.Console.PrintMessage(
            "Task panel: found body references: {}\n{}\n".format(
                items, self.body_selection_widget.references
            )
        )

        # if no solid is added as reference all volume elements are used
        """
        if items == 0:
            msgBox = QtGui.QMessageBox()
            msgBox.setIcon(QtGui.QMessageBox.Question)
            msgBox.setText("Constraint Centrif requires at least one solid")
            msgBox.setWindowTitle("FEM Constraint Centrifuge - Body Selection")
            retryButton = msgBox.addButton(QtGui.QMessageBox.Retry)
            ignoreButton = msgBox.addButton(QtGui.QMessageBox.Ignore)
            msgBox.exec_()

            if msgBox.clickedButton() == retryButton:
                return False
            elif msgBox.clickedButton() == ignoreButton:
                pass
        """

        # check value RotationFrequency
        if self.rotation_frequency == 0:
            msgBox = QtGui.QMessageBox()
            msgBox.setIcon(QtGui.QMessageBox.Question)
            msgBox.setText("Rotational speed is zero")
            msgBox.setWindowTitle("FEM Constraint Centrifuge - Rotational Speed Setting")
            retryButton = msgBox.addButton(QtGui.QMessageBox.Retry)
            ignoreButton = msgBox.addButton(QtGui.QMessageBox.Ignore)
            msgBox.exec_()

            if msgBox.clickedButton() == retryButton:
                return False
            elif msgBox.clickedButton() == ignoreButton:
                pass

        self.obj.RotationFrequency = self.rotation_frequency
        self.obj.RotationAxis = self.axis_selection_widget.references
        self.obj.References = self.body_selection_widget.references
        self.axis_selection_widget.finish_selection()
        self.body_selection_widget.finish_selection()
        return super().accept()

    def reject(self):
        self.axis_selection_widget.finish_selection()
        self.body_selection_widget.finish_selection()
        return super().reject()

    def init_parameter_widget(self):
        self.rotation_frequency = self.obj.RotationFrequency
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_rotation_frequency).bind(
            self.obj, "RotationFrequency"
        )
        self.parameter_widget.qsb_rotation_frequency.setProperty("value", self.rotation_frequency)

    def rotation_frequency_changed(self, base_quantity_value):
        self.rotation_frequency = base_quantity_value



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "CompoundTools._CommandExplodeCompound"
__author__ = "DeepSOIC"
__url__ = "https://www.freecad.org"
__doc__ = (
    "ExplodeCompound: create a bunch of CompoundFilter objects to split a compound into pieces."
)

from .Explode import explodeCompound

import FreeCAD

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtGui
    from PySide import QtCore

    # translation-related code
    try:
        _fromUtf8 = QtCore.QString.fromUtf8
    except Exception:

        def _fromUtf8(s):
            return s

    translate = FreeCAD.Qt.translate


# command class
class _CommandExplodeCompound:
    "Command to explode a compound"

    def GetResources(self):
        return {
            "Pixmap": "Part_ExplodeCompound",
            "MenuText": QtCore.QT_TRANSLATE_NOOP("Part_ExplodeCompound", "Explode Compound"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "Part_ExplodeCompound",
                "Splits up a compound of shapes into separate objects, creating a compound filter for each shape",
            ),
        }

    def Activated(self):
        if len(FreeCADGui.Selection.getSelection()) == 1:
            cmdExplode()
        else:
            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(
                translate("Part_ExplodeCompound", "First select a shape that is a compound.", None)
            )
            mb.setWindowTitle(translate("Part_ExplodeCompound", "Bad Selection", None))
            mb.exec_()

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


if FreeCAD.GuiUp:
    FreeCADGui.addCommand("Part_ExplodeCompound", _CommandExplodeCompound())


def cmdExplode():
    FreeCAD.ActiveDocument.openTransaction("Explode")
    try:
        sel = FreeCADGui.Selection.getSelectionEx()
        if len(sel) != 1:
            raise ValueError(
                "Bad selection",
                "More than one object selected. You have selected {num} objects.".format(
                    num=len(sel)
                ),
            )
        obj = sel[0].Object
        FreeCADGui.addModule("CompoundTools.Explode")
        FreeCADGui.doCommand("input_obj = App.ActiveDocument." + obj.Name)
        FreeCADGui.doCommand("CompoundTools.Explode.explodeCompound(input_obj)")
        FreeCADGui.doCommand("input_obj.ViewObject.hide()")
    except Exception as ex:
        FreeCAD.ActiveDocument.abortTransaction()
        FreeCAD.Console.PrintError("{}\n".format(ex))

    FreeCAD.ActiveDocument.commitTransaction()
    FreeCADGui.doCommand("App.ActiveDocument.recompute()")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the dimension objects.

This includes the `LinearDimension` and `AgularDimension`.
The first one measures a distance between two points or vertices
in an object; it includes radial dimensions of circular arcs.
The second one creates an arc between two straight lines to measure
the angle between both.

To Do
-----
1. Currently the angular dimension does not use linked geometrical
elements, meaning that it cannot update its `Angle` by picking lines or edges
from objects. If fact, `LinkedGeometry` is hidden to prevent the user
from picking any object.

At the moment the user must manually modify `FirstAngle` and `LastAngle`
to obtain a new `Angle`, but since the values are manually entered
the result is not parametrically tied to any actual object in the document.

We introduced a function `measure_two_obj_angles` to calculate
the corresponding parameters from a pair of objects and their edges.
Currently this function is deactivated because we don't consider it
to be ready; it is there for testing purposes only.
This needs to be improved because at the moment it only gives
one possible angle. We should be able to get the four angles
of a two-line intersection. Maybe a new property is required
to indicate the quadrant to choose and display.

2. In general, the `LinkedGeometry` property must be changed in type,
as it does not need to be an `App::PropertyLinkSubList`.
A `LinkSubList` is to select multiple subelements (vertices, edges)
from multiple objects (two lines). However, since we typically measure
a single object, for example, a single line or circle, the subelements
that we can choose must belong to this object only.

Therefore, just like in the case of the `PathArray` class the best property
that could be used is `App::PropertyLinkSub`.
Then in the property editor we will be unable to select more than one object
thus preventing errors of the subelements not matching the measured object.

3. Currently the `LinearDimension` class is able to measure the distance
between two arbitrary vertices in two distinct objects.
For this case `App::PropertyLinkSubList` is in fact the right property
to use, however, neither the `make_dimension` functions
nor the Gui Command are set up to use this type of information.
This has to be done manually by picking the two objects and the two vertices
in the property editor. That is, this functionality is not entirely intuitive,
so it is somewhat 'hidden' from the user.

So, the make function and the Gui Command should be expanded to consider
this case.

Another possibility would be to use one property (LinkSub) for single-object
measurements (linear, radial), and a second property (LinkSubList)
for two-object measurements (linear, angular). This would require adjustments
to the `execute` method to handle both cases properly. It may be necessary
to have another property to control which type to use.

4. The `Support` property is not used at all, so it should be removed.
It is just set at creation time by the `make_dimension` function
but it actually isn't used in the `execute` code.

5. In fact the `DimensionBase` class is not the best base class
than can be used as parent for all dimensions because it defines `Normal`,
`Support`, and `LinkedGeometry`, which aren't used in all cases.
In some of the derived classes, these properties are hidden.

So, together with what is explained in point 3, we probably need to use
a more generic base class, while at the same time improve the way
the link properties are used.
"""
## @package dimension
# \ingroup draftobjects
# \brief Provides the object code for the dimension objects.

## \addtogroup draftobjects
# @{
import math
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import DraftVecUtils
import DraftGeomUtils
import WorkingPlane
from draftobjects.draft_annotation import DraftAnnotation
from draftutils import gui_utils
from draftutils import utils
from draftutils.messages import _log


class DimensionBase(DraftAnnotation):
    """The base objects for dimension objects.

    This class inherits `DraftAnnotation` to define the basic properties
    of all annotation type objects, like a scaling multiplier.

    This class is not used directly, but inherited by all dimension
    objects.
    """

    def set_properties(self, obj):
        """Set basic properties only if they don't exist."""
        properties = obj.PropertiesList

        if "Normal" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "The normal direction of the text " "of the dimension"
            )
            obj.addProperty("App::PropertyVector", "Normal", "Dimension", _tip, locked=True)
            obj.Normal = App.Vector(0, 0, 1)

        # TODO: remove Support property as it is not used at all.
        # It is just set at creation time by the make_dimension function
        # but it is not used.
        if "Support" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "The object measured by this dimension")
            obj.addProperty("App::PropertyLink", "Support", "Dimension", _tip, locked=True)
            obj.Support = None

        if "LinkedGeometry" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The object, and specific subelements "
                "of it,\n"
                "that this dimension "
                "is measuring.\n"
                "\n"
                "There are various possibilities:\n"
                "- An object, and one of its edges.\n"
                "- An object, and two of its vertices.\n"
                "- An arc object, and its edge.",
            )
            obj.addProperty(
                "App::PropertyLinkSubList", "LinkedGeometry", "Dimension", _tip, locked=True
            )
            obj.LinkedGeometry = []

        if "Dimline" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "A point through which the dimension "
                "line, or an extrapolation of it, "
                "will pass.\n"
                "\n"
                "- For linear dimensions, this property "
                "controls how close the dimension line\n"
                "is to the measured object.\n"
                "- For radial dimensions, this controls "
                "the direction of the dimension line\n"
                "that displays the measured radius or "
                "diameter.\n"
                "- For angular dimensions, "
                "this controls the radius of the "
                "dimension arc\n"
                "that displays the measured angle.",
            )
            obj.addProperty(
                "App::PropertyVectorDistance", "Dimline", "Dimension", _tip, locked=True
            )
            obj.Dimline = App.Vector(0, 1, 0)

    def update_properties_0v21(self, obj, vobj):
        """Update view properties."""
        vobj.Proxy.set_text_properties(vobj, vobj.PropertiesList)
        vobj.TextColor = vobj.LineColor
        _log("v0.21, " + obj.Name + ", added view property 'TextColor'")
        _log("v0.21, " + obj.Name + ", renamed 'DisplayMode' options to 'World/Screen'")


class LinearDimension(DimensionBase):
    """The linear dimension object.

    This inherits `DimensionBase` to provide the basic functionality of
    a dimension.

    This linear dimension includes measurements between two vertices,
    but also a radial dimension of a circular edge or arc.
    """

    def __init__(self, obj):
        obj.Proxy = self
        self.Type = "LinearDimension"
        self.set_properties(obj)

    def set_properties(self, obj):
        """Set basic properties only if they don't exist."""
        super().set_properties(obj)

        properties = obj.PropertiesList

        if "Start" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Starting point of the dimension line.\n"
                "\n"
                "If it is a radius dimension it will be "
                "the center of the arc.\n"
                "If it is a diameter dimension "
                "it will be a point that lies "
                "on the arc.",
            )
            obj.addProperty(
                "App::PropertyVectorDistance", "Start", "Linear/radial dimension", _tip, locked=True
            )
            obj.Start = App.Vector(0, 0, 0)

        if "End" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Ending point of the dimension line.\n"
                "\n"
                "If it is a radius or diameter "
                "dimension\n"
                "it will be a point that lies "
                "on the arc.",
            )
            obj.addProperty(
                "App::PropertyVectorDistance", "End", "Linear/radial dimension", _tip, locked=True
            )
            obj.End = App.Vector(1, 0, 0)

        if "Direction" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The direction of the dimension line.\n"
                "If this remains '(0,0,0)', "
                "the direction will be calculated "
                "automatically.",
            )
            obj.addProperty(
                "App::PropertyVector", "Direction", "Linear/radial dimension", _tip, locked=True
            )

        if "Distance" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The value of the measurement.\n"
                "\n"
                "This property is read-only because "
                "the value is calculated\n"
                "from the 'Start' and 'End' properties.\n"
                "\n"
                "If the 'Linked Geometry' "
                "is an arc or circle, this 'Distance'\n"
                "is the radius or diameter, depending "
                "on the 'Diameter' property.",
            )
            obj.addProperty(
                "App::PropertyLength", "Distance", "Linear/radial dimension", _tip, locked=True
            )
            obj.Distance = 0

        if "Diameter" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "When measuring circular arcs, "
                "it determines whether to display\n"
                "the radius or the diameter value",
            )
            obj.addProperty("App::PropertyBool", "Diameter", "Radial dimension", _tip, locked=True)
            obj.Diameter = False

    def onDocumentRestored(self, obj):
        """Execute code when the document is restored."""
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(
            obj, vp_module="view_dimension", vp_class="ViewProviderLinearDimension"
        )

        vobj = getattr(obj, "ViewObject", None)
        if vobj is None:
            return

        if not hasattr(vobj, "TextColor"):
            self.update_properties_0v21(obj, vobj)

    def loads(self, state):
        self.Type = "LinearDimension"

    def onChanged(self, obj, prop):
        """Execute when a property is changed.

        It just sets some properties to be read-only or hidden,
        as they aren't used.
        """
        if hasattr(obj, "Distance"):
            obj.setPropertyStatus("Distance", "ReadOnly")

        # if hasattr(obj, "Normal"):
        #    obj.setPropertyStatus('Normal', 'Hidden')
        if hasattr(obj, "Support"):
            obj.setPropertyStatus("Support", "Hidden")

    def transform(self, obj, pla):
        """Transform the object by applying a placement."""
        obj.Start = pla.multVec(obj.Start)
        obj.End = pla.multVec(obj.End)
        obj.Dimline = pla.multVec(obj.Dimline)
        obj.Normal = pla.Rotation.multVec(obj.Normal)
        obj.Direction = pla.Rotation.multVec(obj.Direction)

    def execute(self, obj):
        """Execute when the object is created or recomputed.

        Set start point and end point according to the linked geometry
        and the number of subelements.

        If it has one subelement, we assume a straight edge or a circular edge.
        If it has two subelements, we assume a straight edge (two vertices).
        """
        if obj.LinkedGeometry:
            if len(obj.LinkedGeometry) == 1:
                linked_obj = obj.LinkedGeometry[0][0]
                sub_list = obj.LinkedGeometry[0][1]

                if len(sub_list) == 1:
                    # If it has one subelement, we assume an edge
                    # that can be a straight line, or a circular edge
                    subelement = sub_list[0]
                    (obj.Start, obj.End) = measure_one_obj_edge(
                        linked_obj, subelement, obj.Dimline, obj.Diameter
                    )
                elif len(sub_list) == 2:
                    # If it has two subelements, we assume a straight edge
                    # that is measured by two vertices
                    (obj.Start, obj.End) = measure_one_obj_vertices(linked_obj, sub_list)

            elif len(obj.LinkedGeometry) == 2:
                # If the list has two objects, it measures the distance
                # between the two vertices in those two objects
                (obj.Start, obj.End) = measure_two_objects(
                    obj.LinkedGeometry[0], obj.LinkedGeometry[1]
                )

        # Update the distance property by comparing the floats
        # with the precision
        net_length = (obj.Start.sub(obj.End)).Length
        rounded_1 = round(obj.Distance.Value, utils.precision())
        rounded_2 = round(net_length, utils.precision())

        if rounded_1 != rounded_2:
            obj.Distance = net_length

        # The lines and text are created in the viewprovider, so we should
        # update it whenever the object is recomputed
        if App.GuiUp and obj.ViewObject:
            obj.ViewObject.update()


def measure_one_obj_edge(obj, subelement, dim_point, diameter=False):
    """Measure one object with one subelement, a straight or circular edge.

    Parameters
    ----------
    obj: Part::Feature
        The object that is measured.

    subelement: str
        The subelement that is measured, for example, `'Edge1'`.

    dim_line: Base::Vector3
        A point through which the dimension goes through.
    """
    start = App.Vector()
    end = App.Vector()

    if "Edge" in subelement:
        n = int(subelement[4:]) - 1
        edge = obj.Shape.Edges[n]

        if DraftGeomUtils.geomType(edge) == "Line":
            start = edge.Vertexes[0].Point
            end = edge.Vertexes[-1].Point
        elif DraftGeomUtils.geomType(edge) == "Circle":
            center = edge.Curve.Center
            radius = edge.Curve.Radius
            axis = edge.Curve.Axis
            dim_line = dim_point.sub(center)

            # The ray is projected to the plane on which the circle lies,
            # but if the projection is not successful, try in the other planes
            ray = dim_line.projectToPlane(App.Vector(0, 0, 0), axis)

            if ray.Length == 0:
                ray = axis.cross(App.Vector(1, 0, 0))
                if ray.Length == 0:
                    ray = axis.cross(App.Vector(0, 1, 0))

            # The ray is made as large as the arc's radius
            # and optionally the diameter
            ray = DraftVecUtils.scaleTo(ray, radius)

            if diameter:
                # The start and end points lie on the arc
                start = center.add(ray.negative())
                end = center.add(ray)
            else:
                # The start is th center and the end lies on the arc
                start = center
                end = center.add(ray)

    return start, end


def measure_one_obj_vertices(obj, subelements):
    """Measure two vertices in the same object."""
    start = App.Vector()
    end = App.Vector()

    subelement1 = subelements[0]
    subelement2 = subelements[1]

    if "Vertex" in subelement1 and "Vertex" in subelement2:
        n1 = int(subelement1[6:]) - 1
        n2 = int(subelement2[6:]) - 1
        start = obj.Shape.Vertexes[n1].Point
        end = obj.Shape.Vertexes[n2].Point

    return start, end


def measure_two_objects(link_sub_1, link_sub_2):
    """Measure two vertices from two different objects.

    Parameters
    ----------
    link_sub_1: tuple
        A tuple containing one object and a list of subelement strings,
        which may be empty. Only the first subelement is considered, which
        must be a vertex.
        ::
            link_sub_1 = (obj1, ['VertexN', ...])

    link_sub_2: tuple
        Same.
    """
    start = App.Vector()
    end = App.Vector()

    obj1 = link_sub_1[0]
    lsub1 = link_sub_1[1]

    obj2 = link_sub_2[0]
    lsub2 = link_sub_2[1]

    # The subelement list may be empty so we test it first
    # and pick only the first item
    if lsub1 and lsub2:
        subelement1 = lsub1[0]
        subelement2 = lsub2[0]

        if "Vertex" in subelement1 and "Vertex" in subelement2:
            n1 = int(subelement1[6:]) - 1
            n2 = int(subelement2[6:]) - 1
            start = obj1.Shape.Vertexes[n1].Point
            end = obj2.Shape.Vertexes[n2].Point

    return start, end


# Alias for compatibility with v0.18 and earlier
_Dimension = LinearDimension


class AngularDimension(DimensionBase):
    """The angular dimension object.

    This inherits `DimensionBase` to provide the basic functionality of
    a dimension.
    """

    def __init__(self, obj):
        obj.Proxy = self
        self.Type = "AngularDimension"
        self.set_properties(obj)

    def set_properties(self, obj):
        """Set basic properties only if they don't exist."""
        super().set_properties(obj)

        properties = obj.PropertiesList

        if "FirstAngle" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Starting angle of the dimension line "
                "(circular arc).\n"
                "The arc is drawn counter-clockwise.",
            )
            obj.addProperty(
                "App::PropertyAngle", "FirstAngle", "Angular dimension", _tip, locked=True
            )
            obj.FirstAngle = 0

        if "LastAngle" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Ending angle of the dimension line "
                "(circular arc).\n"
                "The arc is drawn counter-clockwise.",
            )
            obj.addProperty(
                "App::PropertyAngle", "LastAngle", "Angular dimension", _tip, locked=True
            )
            obj.LastAngle = 90

        if "Center" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The center point of the dimension "
                "line, which is a circular arc.\n"
                "\n"
                "This is normally the point where two "
                "line segments, or their extensions\n"
                "intersect, resulting in the "
                "measured 'Angle' between them.",
            )
            obj.addProperty(
                "App::PropertyVectorDistance", "Center", "Angular dimension", _tip, locked=True
            )
            obj.Center = App.Vector(0, 0, 0)

        if "Angle" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The value of the measurement.\n"
                "\n"
                "This property is read-only because "
                "the value is calculated from\n"
                "the 'First Angle' and "
                "'Last Angle' properties.",
            )
            obj.addProperty("App::PropertyAngle", "Angle", "Angular dimension", _tip, locked=True)
            obj.Angle = 0

    def onDocumentRestored(self, obj):
        """Execute code when the document is restored."""
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(
            obj, vp_module="view_dimension", vp_class="ViewProviderAngularDimension"
        )

        vobj = getattr(obj, "ViewObject", None)
        if vobj is None:
            return

        if not hasattr(vobj, "TextColor"):
            self.update_properties_0v21(obj, vobj)

    def loads(self, state):
        self.Type = "AngularDimension"

    def transform(self, obj, pla):
        """Transform the object by applying a placement."""
        import Part

        new_normal = pla.Rotation.multVec(obj.Normal)
        c_old = Part.makeCircle(1, App.Vector(), obj.Normal)
        c_tra = c_old.transformShape((pla * c_old.Placement).Rotation.Matrix)
        c_new = Part.makeCircle(1, App.Vector(), new_normal)
        delta_angle = math.degrees(c_new.Curve.parameter(c_tra.firstVertex().Point))
        first_angle = (obj.FirstAngle.Value + delta_angle) % 360
        last_angle = (obj.LastAngle.Value + delta_angle) % 360

        obj.Center = pla.multVec(obj.Center)
        obj.Dimline = pla.multVec(obj.Dimline)
        obj.Normal = new_normal
        obj.FirstAngle = first_angle
        obj.LastAngle = last_angle

    def execute(self, obj):
        """Execute when the object is created or recomputed.

        Nothing is actually done here, except update the viewprovider,
        as the lines and text are created in the viewprovider.
        """
        # TODO: introduce the calculation of 'Angle' by taking a pair of
        # objects and edges in the 'LinkedGeometry' property.
        #
        # We introduced a function `measure_two_obj_angles` to calculate
        # the corresponding parameters from a pair of objects and their edges.
        # Currently this function is deactivated because we don't consider it
        # to be ready; it is there for testing purposes only.
        #
        # if obj.LinkedGeometry and len(obj.LinkedGeometry) == 2:
        #     (obj.FirstAngle,
        #      obj.LastAngle) = measure_two_obj_angles(obj.LinkedGeometry[0],
        #                                              obj.LinkedGeometry[1])

        # TODO: move the calculation of 'Angle' from the viewprovider
        # to this object class.
        # The viewprovider should modify visual properties only, not real
        # properties. It can react to real properties by using the 'updateData'
        # method.
        if App.GuiUp and obj.ViewObject:
            obj.ViewObject.update()

    def onChanged(self, obj, prop):
        """Execute when a property is changed.

        It just sets some properties to be read-only or hidden,
        as they aren't used.
        """
        if hasattr(obj, "Angle"):
            obj.setPropertyStatus("Angle", "ReadOnly")

        if hasattr(obj, "Normal"):
            obj.setPropertyStatus("Normal", "Hidden")
        if hasattr(obj, "Support"):
            obj.setPropertyStatus("Support", "Hidden")
        if hasattr(obj, "LinkedGeometry"):
            obj.setPropertyStatus("LinkedGeometry", "Hidden")


def measure_two_obj_angles(link_sub_1, link_sub_2):
    """Measure two edges from two different objects to measure the angle.

    This function is a prototype because it does not determine all possible
    starting and ending angles that could be used to draw the dimension line,
    which is a circular arc.

    Parameters
    ----------
    link_sub_1: tuple
        A tuple containing one object and a list of subelement strings,
        which may be empty. Only the first subelement is considered, which
        must be an edge.
        ::
            link_sub_1 = (obj1, ['EdgeN', ...])

    link_sub_2: tuple
        Same.
    """
    start = 0
    end = 0

    obj1 = link_sub_1[0]
    lsub1 = link_sub_1[1]

    obj2 = link_sub_2[0]
    lsub2 = link_sub_2[1]

    # The subelement list may be empty so we test it first
    # and pick only the first item
    if lsub1 and lsub2:
        subelement1 = lsub1[0]
        subelement2 = lsub2[0]

        if "Edge" in subelement1 and "Edge" in subelement2:
            n1 = int(subelement1[4:]) - 1
            n2 = int(subelement2[4:]) - 1
            start = obj1.Shape.Edges[n1].Curve.Direction
            end = obj2.Shape.Edges[n2].Curve.Direction

            # We get the angle from the U axis of the working plane
            wp = WorkingPlane.get_working_plane(update=False)
            start_r = DraftVecUtils.angle(start, wp.u)
            end_r = DraftVecUtils.angle(end, wp.u)
            start = math.degrees(start_r)
            end = math.degrees(end_r)

            # We make the angle positive because when tested, some errors
            # were produced in the code that calculates the 'Angle'.
            # This code is actually inside the viewprovider.
            if start < 0:
                start = abs(start)
            if end < 0:
                end = abs(end)

    return start, end


# Alias for compatibility with v0.18 and earlier
_AngularDimension = AngularDimension

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import Path
import CAMTests.PathTestUtils as PathTestUtils


class TestPathPreferences(PathTestUtils.PathTestBase):
    def test00(self):
        """There is at least one search path."""

        paths = Path.Preferences.searchPaths()
        self.assertGreater(len(paths), 0)

    def test01(self):
        """Path/Post is part of the posts search path."""
        paths = Path.Preferences.searchPathsPost()
        self.assertEqual(len([p for p in paths if p.endswith("/Path/Post/")]), 1)

    def test02(self):
        """Path/Post/scripts is part of the posts search path."""
        paths = Path.Preferences.searchPathsPost()
        self.assertEqual(len([p for p in paths if p.endswith("/Path/Post/scripts/")]), 1)

    def test03(self):
        """Available post processors include linuxcnc, grbl and opensbp."""
        posts = Path.Preferences.allAvailablePostProcessors()
        self.assertTrue("linuxcnc" in posts)
        self.assertTrue("grbl" in posts)
        self.assertTrue("opensbp" in posts)

    def test10(self):
        """Default paths for tools are resolved correctly"""

        self.assertEqual(
            Path.Preferences.getDefaultAssetPath().parts[-1],
            "CamAssets",
            str(Path.Preferences.getDefaultAssetPath()),
        )
        self.assertEqual(
            Path.Preferences.getBuiltinAssetPath().parts[-2:],
            ("CAM", "Tools"),
            str(Path.Preferences.getBuiltinAssetPath()),
        )
        self.assertEqual(
            Path.Preferences.getBuiltinShapePath().parts[-3:],
            ("CAM", "Tools", "Shape"),
            str(Path.Preferences.getBuiltinShapePath()),
        )
        self.assertEqual(Path.Preferences.getToolBitPath().name, "Bit")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import uuid
import json
import FreeCAD
from FreeCAD import Base
from typing import Optional, Union, Mapping, List
from ...assets import Asset, AssetUri, AssetSerializer


class Machine(Asset):
    """Represents a machine with various operational parameters."""

    asset_type: str = "machine"
    API_VERSION = 1

    UNIT_CONVERSIONS = {
        "hp": 745.7,  # hp to W
        "in-lbf": 0.112985,  # in-lbf to N*m
        "inch/min": 25.4,  # inch/min to mm/min
        "rpm": 1.0 / 60.0,  # rpm to 1/s
        "kW": 1000.0,  # kW to W
        "Nm": 1.0,  # Nm to N*m
        "mm/min": 1.0,  # mm/min to mm/min
    }

    def __init__(
        self,
        label: str = "Machine",
        max_power: Union[int, float, FreeCAD.Units.Quantity] = 2,
        min_rpm: Union[int, float, FreeCAD.Units.Quantity] = 3000,
        max_rpm: Union[int, float, FreeCAD.Units.Quantity] = 60000,
        max_torque: Optional[Union[int, float, FreeCAD.Units.Quantity]] = None,
        peak_torque_rpm: Optional[Union[int, float, FreeCAD.Units.Quantity]] = None,
        min_feed: Union[int, float, FreeCAD.Units.Quantity] = 1,
        max_feed: Union[int, float, FreeCAD.Units.Quantity] = 2000,
        id: Optional[str] = None,
    ) -> None:
        """
        Initializes a Machine object.

        Args:
            label: The label of the machine.
            max_power: The maximum power of the machine (kW or Quantity).
            min_rpm: The minimum RPM of the machine (RPM or Quantity).
            max_rpm: The maximum RPM of the machine (RPM or Quantity).
            max_torque: The maximum torque of the machine (Nm or Quantity).
            peak_torque_rpm: The RPM at which peak torque is achieved
                             (RPM or Quantity).
            min_feed: The minimum feed rate of the machine
                      (mm/min or Quantity).
            max_feed: The maximum feed rate of the machine
                      (mm/min or Quantity).
            id: The unique identifier of the machine.
        """
        self.id = id or str(uuid.uuid1())
        self._label = label

        # Initialize max_power (W)
        if isinstance(max_power, FreeCAD.Units.Quantity):
            self._max_power = max_power.getValueAs("W").Value
        elif isinstance(max_power, (int, float)):
            self._max_power = max_power * self.UNIT_CONVERSIONS["kW"]
        else:
            self._max_power = 2000.0

        # Initialize min_rpm (1/s)
        if isinstance(min_rpm, FreeCAD.Units.Quantity):
            try:
                self._min_rpm = min_rpm.getValueAs("1/s").Value
            except (Base.ParserError, ValueError):
                self._min_rpm = min_rpm.Value * self.UNIT_CONVERSIONS["rpm"]
        elif isinstance(min_rpm, (int, float)):
            self._min_rpm = min_rpm * self.UNIT_CONVERSIONS["rpm"]
        else:
            self._min_rpm = 3000 * self.UNIT_CONVERSIONS["rpm"]

        # Initialize max_rpm (1/s)
        if isinstance(max_rpm, FreeCAD.Units.Quantity):
            try:
                self._max_rpm = max_rpm.getValueAs("1/s").Value
            except (Base.ParserError, ValueError):
                self._max_rpm = max_rpm.Value * self.UNIT_CONVERSIONS["rpm"]
        elif isinstance(max_rpm, (int, float)):
            self._max_rpm = max_rpm * self.UNIT_CONVERSIONS["rpm"]
        else:
            self._max_rpm = 60000 * self.UNIT_CONVERSIONS["rpm"]

        # Initialize min_feed (mm/min)
        if isinstance(min_feed, FreeCAD.Units.Quantity):
            self._min_feed = min_feed.getValueAs("mm/min").Value
        elif isinstance(min_feed, (int, float)):
            self._min_feed = min_feed
        else:
            self._min_feed = 1.0

        # Initialize max_feed (mm/min)
        if isinstance(max_feed, FreeCAD.Units.Quantity):
            self._max_feed = max_feed.getValueAs("mm/min").Value
        elif isinstance(max_feed, (int, float)):
            self._max_feed = max_feed
        else:
            self._max_feed = 2000.0

        # Initialize peak_torque_rpm (1/s)
        if isinstance(peak_torque_rpm, FreeCAD.Units.Quantity):
            try:
                self._peak_torque_rpm = peak_torque_rpm.getValueAs("1/s").Value
            except (Base.ParserError, ValueError):
                self._peak_torque_rpm = peak_torque_rpm.Value * self.UNIT_CONVERSIONS["rpm"]
        elif isinstance(peak_torque_rpm, (int, float)):
            self._peak_torque_rpm = peak_torque_rpm * self.UNIT_CONVERSIONS["rpm"]
        else:
            self._peak_torque_rpm = self._max_rpm / 3

        # Initialize max_torque (N*m)
        if isinstance(max_torque, FreeCAD.Units.Quantity):
            self._max_torque = max_torque.getValueAs("Nm").Value
        elif isinstance(max_torque, (int, float)):
            self._max_torque = max_torque
        else:
            # Convert 1/s to rpm
            peak_rpm_for_calc = self._peak_torque_rpm * 60
            self._max_torque = (
                self._max_power * 9.5488 / peak_rpm_for_calc if peak_rpm_for_calc else float("inf")
            )

    def get_id(self) -> str:
        """Returns the unique identifier for the Machine instance."""
        return self.id

    def to_dict(self) -> dict:
        """Returns a dictionary representation of the Machine."""
        return {
            "version": self.API_VERSION,
            "id": self.id,
            "label": self.label,
            "max_power": self._max_power,  # W
            "min_rpm": self._min_rpm,  # 1/s
            "max_rpm": self._max_rpm,  # 1/s
            "max_torque": self._max_torque,  # Nm
            "peak_torque_rpm": self._peak_torque_rpm,  # 1/s
            "min_feed": self._min_feed,  # mm/min
            "max_feed": self._max_feed,  # mm/min
        }

    def to_bytes(self, serializer: AssetSerializer) -> bytes:
        """Serializes the Machine object to bytes using to_dict."""
        data_dict = self.to_dict()
        json_str = json.dumps(data_dict)
        return json_str.encode("utf-8")

    @classmethod
    def from_dict(cls, data_dict: dict, id: str) -> "Machine":
        """Creates a Machine instance from a dictionary."""
        machine = cls(
            label=data_dict.get("label", "Machine"),
            max_power=data_dict.get("max_power", 2000.0),  # W
            min_rpm=data_dict.get("min_rpm", 3000 * cls.UNIT_CONVERSIONS["rpm"]),  # 1/s
            max_rpm=data_dict.get("max_rpm", 60000 * cls.UNIT_CONVERSIONS["rpm"]),  # 1/s
            max_torque=data_dict.get("max_torque", None),  # Nm
            peak_torque_rpm=data_dict.get("peak_torque_rpm", None),  # 1/s
            min_feed=data_dict.get("min_feed", 1.0),  # mm/min
            max_feed=data_dict.get("max_feed", 2000.0),  # mm/min
            id=id,
        )
        return machine

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
    ) -> "Machine":
        """
        Deserializes bytes into a Machine instance using from_dict.
        """
        # If dependencies is None, it's fine as Machine doesn't use it.
        data_dict = json.loads(data.decode("utf-8"))
        return cls.from_dict(data_dict, id)

    @classmethod
    def dependencies(cls, data: bytes) -> List[AssetUri]:
        """Returns a list of AssetUri dependencies parsed from the serialized data."""
        return []  # Machine has no dependencies

    @property
    def max_power(self) -> FreeCAD.Units.Quantity:
        return FreeCAD.Units.Quantity(self._max_power, "W")

    @property
    def min_rpm(self) -> FreeCAD.Units.Quantity:
        return FreeCAD.Units.Quantity(self._min_rpm, "1/s")

    @property
    def max_rpm(self) -> FreeCAD.Units.Quantity:
        return FreeCAD.Units.Quantity(self._max_rpm, "1/s")

    @property
    def max_torque(self) -> FreeCAD.Units.Quantity:
        return FreeCAD.Units.Quantity(self._max_torque, "Nm")

    @property
    def peak_torque_rpm(self) -> FreeCAD.Units.Quantity:
        return FreeCAD.Units.Quantity(self._peak_torque_rpm, "1/s")

    @property
    def min_feed(self) -> FreeCAD.Units.Quantity:
        return FreeCAD.Units.Quantity(self._min_feed, "mm/min")

    @property
    def max_feed(self) -> FreeCAD.Units.Quantity:
        return FreeCAD.Units.Quantity(self._max_feed, "mm/min")

    @property
    def label(self) -> str:
        return self._label

    @label.setter
    def label(self, label: str) -> None:
        self._label = label

    def get_min_rpm_value(self) -> float:
        """Helper method to get minimum RPM value for display/testing."""
        return self._min_rpm * 60

    def get_max_rpm_value(self) -> float:
        """Helper method to get maximum RPM value for display/testing."""
        return self._max_rpm * 60

    def get_peak_torque_rpm_value(self) -> float:
        """Helper method to get peak torque RPM value for display/testing."""
        return self._peak_torque_rpm * 60

    def validate(self) -> None:
        """Validates the machine parameters."""
        if not self.label:
            raise AttributeError("Machine name is required")
        if self._peak_torque_rpm > self._max_rpm:
            err = ("Peak Torque RPM {ptrpm:.2f} must be less than max RPM " "{max_rpm:.2f}").format(
                ptrpm=self._peak_torque_rpm * 60, max_rpm=self._max_rpm * 60
            )
            raise AttributeError(err)
        if self._max_rpm <= self._min_rpm:
            raise AttributeError("Max RPM must be larger than min RPM")
        if self._max_feed <= self._min_feed:
            raise AttributeError("Max feed must be larger than min feed")

    def get_torque_at_rpm(self, rpm: Union[int, float, FreeCAD.Units.Quantity]) -> float:
        """
        Calculates the torque at a given RPM.

        Args:
            rpm: The RPM value (int, float, or Quantity).

        Returns:
            The torque at the given RPM in Nm.
        """
        if isinstance(rpm, FreeCAD.Units.Quantity):
            try:
                rpm_hz = rpm.getValueAs("1/s").Value
            except (Base.ParserError, ValueError):
                rpm_hz = rpm.Value * self.UNIT_CONVERSIONS["rpm"]
        else:
            rpm_hz = rpm * self.UNIT_CONVERSIONS["rpm"]
        max_torque_nm = self._max_torque
        peak_torque_rpm_hz = self._peak_torque_rpm
        peak_rpm_for_calc = peak_torque_rpm_hz * 60
        rpm_for_calc = rpm_hz * 60
        torque_at_current_rpm = (
            self._max_power * 9.5488 / rpm_for_calc if rpm_for_calc else float("inf")
        )
        if rpm_for_calc <= peak_rpm_for_calc:
            torque_at_current_rpm = (
                max_torque_nm / peak_rpm_for_calc * rpm_for_calc
                if peak_rpm_for_calc
                else float("inf")
            )
        return min(max_torque_nm, torque_at_current_rpm)

    def set_max_power(self, power: Union[int, float], unit: Optional[str] = None) -> None:
        """Sets the maximum power of the machine."""
        unit = unit or "kW"
        if unit in self.UNIT_CONVERSIONS:
            power_value = power * self.UNIT_CONVERSIONS[unit]
        else:
            power_value = FreeCAD.Units.Quantity(power, unit).getValueAs("W").Value
        self._max_power = power_value
        if self._max_power <= 0:
            raise AttributeError("Max power must be positive")

    def set_min_rpm(self, min_rpm: Union[int, float, FreeCAD.Units.Quantity]) -> None:
        """Sets the minimum RPM of the machine."""
        if isinstance(min_rpm, FreeCAD.Units.Quantity):
            try:
                min_rpm_value = min_rpm.getValueAs("1/s").Value
            except (Base.ParserError, ValueError):
                min_rpm_value = min_rpm.Value * self.UNIT_CONVERSIONS["rpm"]
        else:
            min_rpm_value = min_rpm * self.UNIT_CONVERSIONS["rpm"]
        self._min_rpm = min_rpm_value
        if self._min_rpm < 0:
            raise AttributeError("Min RPM cannot be negative")
        if self._min_rpm >= self._max_rpm:
            self._max_rpm = min_rpm_value + 1.0 / 60.0

    def set_max_rpm(self, max_rpm: Union[int, float, FreeCAD.Units.Quantity]) -> None:
        """Sets the maximum RPM of the machine."""
        if isinstance(max_rpm, FreeCAD.Units.Quantity):
            try:
                max_rpm_value = max_rpm.getValueAs("1/s").Value
            except (Base.ParserError, ValueError):
                max_rpm_value = max_rpm.Value * self.UNIT_CONVERSIONS["rpm"]
        else:
            max_rpm_value = max_rpm * self.UNIT_CONVERSIONS["rpm"]
        self._max_rpm = max_rpm_value
        if self._max_rpm <= 0:
            raise AttributeError("Max RPM must be positive")
        if self._max_rpm <= self._min_rpm:
            self._min_rpm = max(0, max_rpm_value - 1.0 / 60.0)

    def set_min_feed(
        self,
        min_feed: Union[int, float, FreeCAD.Units.Quantity],
        unit: Optional[str] = None,
    ) -> None:
        """Sets the minimum feed rate of the machine."""
        unit = unit or "mm/min"
        if unit in self.UNIT_CONVERSIONS:
            min_feed_value = min_feed * self.UNIT_CONVERSIONS[unit]
        else:
            min_feed_value = FreeCAD.Units.Quantity(min_feed, unit).getValueAs("mm/min").Value
        self._min_feed = min_feed_value
        if self._min_feed < 0:
            raise AttributeError("Min feed cannot be negative")
        if self._min_feed >= self._max_feed:
            self._max_feed = min_feed_value + 1.0

    def set_max_feed(
        self,
        max_feed: Union[int, float, FreeCAD.Units.Quantity],
        unit: Optional[str] = None,
    ) -> None:
        """Sets the maximum feed rate of the machine."""
        unit = unit or "mm/min"
        if unit in self.UNIT_CONVERSIONS:
            max_feed_value = max_feed * self.UNIT_CONVERSIONS[unit]
        else:
            max_feed_value = FreeCAD.Units.Quantity(max_feed, unit).getValueAs("mm/min").Value
        self._max_feed = max_feed_value
        if self._max_feed <= 0:
            raise AttributeError("Max feed must be positive")
        if self._max_feed <= self._min_feed:
            self._min_feed = max(0, max_feed_value - 1.0)

    def set_peak_torque_rpm(
        self, peak_torque_rpm: Union[int, float, FreeCAD.Units.Quantity]
    ) -> None:
        """Sets the peak torque RPM of the machine."""
        if isinstance(peak_torque_rpm, FreeCAD.Units.Quantity):
            try:
                peak_torque_rpm_value = peak_torque_rpm.getValueAs("1/s").Value
            except (Base.ParserError, ValueError):
                peak_torque_rpm_value = peak_torque_rpm.Value * self.UNIT_CONVERSIONS["rpm"]
        else:
            peak_torque_rpm_value = peak_torque_rpm * self.UNIT_CONVERSIONS["rpm"]
        self._peak_torque_rpm = peak_torque_rpm_value
        if self._peak_torque_rpm < 0:
            raise AttributeError("Peak torque RPM cannot be negative")

    def set_max_torque(
        self,
        max_torque: Union[int, float, FreeCAD.Units.Quantity],
        unit: Optional[str] = None,
    ) -> None:
        """Sets the maximum torque of the machine."""
        unit = unit or "Nm"
        if unit in self.UNIT_CONVERSIONS:
            max_torque_value = max_torque * self.UNIT_CONVERSIONS[unit]
        else:
            max_torque_value = FreeCAD.Units.Quantity(max_torque, unit).getValueAs("Nm").Value
        self._max_torque = max_torque_value
        if self._max_torque <= 0:
            raise AttributeError("Max torque must be positive")

    def dump(self, do_print: bool = True) -> Optional[str]:
        """
        Dumps machine information to console or returns it as a string.

        Args:
            do_print: If True, prints the information to the console.
                      If False, returns the information as a string.

        Returns:
            A formatted string containing machine information if do_print is
            False, otherwise None.
        """
        min_rpm_value = self._min_rpm * 60
        max_rpm_value = self._max_rpm * 60
        peak_torque_rpm_value = self._peak_torque_rpm * 60

        output = ""
        output += f"Machine {self.label}:\n"
        output += f"  Max power: {self._max_power:.2f} W\n"
        output += f"  RPM: {min_rpm_value:.2f} RPM - {max_rpm_value:.2f} RPM\n"
        output += f"  Feed: {self.min_feed.UserString} - " f"{self.max_feed.UserString}\n"
        output += (
            f"  Peak torque: {self._max_torque:.2f} Nm at " f"{peak_torque_rpm_value:.2f} RPM\n"
        )
        output += f"  Max_torque: {self._max_torque} Nm\n"

        if do_print:
            print(output)
        return output



# SPDX-License-Identifier: LGPL-2.1-or-later

# /***************************************************************************
# *   Copyright (c) 2016 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

__title__ = "BOPTools.JoinFeatures module"
__author__ = "DeepSOIC"
__url__ = "https://www.freecad.org"
__doc__ = "Implementation of document objects (features) for connect, ebmed and cutout operations."

from . import JoinAPI
import FreeCAD
import Part

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtCore, QtGui

    # -------------------------- common stuff -------------------------------------

    # -------------------------- translation-related code -------------------------
    # Thanks, yorik! (see forum thread "A new Part tool is being born... JoinFeatures!"
    # https://forum.freecad.org/viewtopic.php?f=22&t=11112&start=30#p90239 )

    try:
        _fromUtf8 = QtCore.QString.fromUtf8
    except Exception:

        def _fromUtf8(s):
            return s

    translate = FreeCAD.Qt.translate
# --------------------------/translation-related code -------------------------


def getParamRefine():
    return FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Part/Boolean").GetBool(
        "RefineModel"
    )


def cmdCreateJoinFeature(name, mode):
    """cmdCreateJoinFeature(name, mode): generalized implementation of GUI commands."""
    sel = FreeCADGui.Selection.getSelectionEx()

    FreeCAD.ActiveDocument.openTransaction("Create " + mode)
    FreeCADGui.addModule("BOPTools.JoinFeatures")
    FreeCADGui.doCommand(
        "j = BOPTools.JoinFeatures.make{mode}(name='{name}')".format(mode=mode, name=name)
    )
    if mode == "Embed" or mode == "Cutout":
        FreeCADGui.doCommand("j.Base = App.ActiveDocument." + sel[0].Object.Name)
        FreeCADGui.doCommand("j.Tool = App.ActiveDocument." + sel[1].Object.Name)
    elif mode == "Connect":
        FreeCADGui.doCommand(
            "j.Objects = {sel}".format(
                sel="[" + ", ".join(["App.ActiveDocument." + so.Object.Name for so in sel]) + "]"
            )
        )
    else:
        raise ValueError("cmdCreateJoinFeature: Unexpected mode {mode}".format(mode=repr(mode)))

    try:
        FreeCADGui.doCommand("j.Proxy.execute(j)")
        FreeCADGui.doCommand("j.purgeTouched()")
    except Exception as err:
        mb = QtGui.QMessageBox()
        mb.setIcon(mb.Icon.Warning)
        error_text1 = translate("Part_JoinFeatures", "Computing the result failed with an error:")
        error_text2 = translate(
            "Part_JoinFeatures",
            "Click 'Continue' to create the feature anyway, or 'Abort' to cancel.",
        )
        mb.setText(error_text1 + "\n\n" + str(err) + "\n\n" + error_text2)
        mb.setWindowTitle(translate("Part_JoinFeatures", "Bad Selection", None))
        btnAbort = mb.addButton(QtGui.QMessageBox.StandardButton.Abort)
        btnOK = mb.addButton(
            translate("Part_JoinFeatures", "Continue", None),
            QtGui.QMessageBox.ButtonRole.ActionRole,
        )
        mb.setDefaultButton(btnOK)
        mb.exec_()

        if mb.clickedButton() is btnAbort:
            FreeCAD.ActiveDocument.abortTransaction()
            return

    FreeCADGui.doCommand(
        "for obj in j.ViewObject.Proxy.claimChildren():\n" "    obj.ViewObject.hide()"
    )

    FreeCAD.ActiveDocument.commitTransaction()


def getIconPath(icon_dot_svg):
    return icon_dot_svg


# -------------------------- /common stuff ------------------------------------

# -------------------------- Connect ------------------------------------------


def makeConnect(name):
    """makeConnect(name): makes an Connect object."""
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", name)
    FeatureConnect(obj)
    if FreeCAD.GuiUp:
        ViewProviderConnect(obj.ViewObject)
    return obj


class FeatureConnect:
    """The PartJoinFeature object."""

    def __init__(self, obj):
        obj.addProperty(
            "App::PropertyLinkList", "Objects", "Connect", "Object to be connected.", locked=True
        )
        obj.addProperty(
            "App::PropertyBool",
            "Refine",
            "Connect",
            "True = refine resulting shape. False = output as is.",
            locked=True,
        )
        obj.Refine = getParamRefine()
        obj.addProperty(
            "App::PropertyLength",
            "Tolerance",
            "Connect",
            "Tolerance when intersecting (fuzzy value). "
            "In addition to tolerances of the shapes.",
            locked=True,
        )

        obj.Proxy = self
        self.Type = "FeatureConnect"

    def execute(self, selfobj):
        rst = JoinAPI.connect([obj.Shape for obj in selfobj.Objects], selfobj.Tolerance)
        if selfobj.Refine:
            rst = rst.removeSplitter()
        selfobj.Shape = rst


class ViewProviderConnect:
    """A View Provider for the Part Connect feature."""

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/booleans/Part_JoinConnect.svg"

    def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def claimChildren(self):
        return self.Object.Objects

    def onDelete(self, feature, subelements):
        try:
            for obj in self.claimChildren():
                obj.ViewObject.show()
        except Exception as err:
            FreeCAD.Console.PrintError("Error in onDelete: " + str(err))
        return True

    def canDragObjects(self):
        return True

    def canDropObjects(self):
        return True

    def canDragObject(self, dragged_object):
        return True

    def canDropObject(self, incoming_object):
        return hasattr(incoming_object, "Shape")

    def dragObject(self, selfvp, dragged_object):
        objs = self.Object.Objects
        objs.remove(dragged_object)
        self.Object.Objects = objs

    def dropObject(self, selfvp, incoming_object):
        self.Object.Objects = self.Object.Objects + [incoming_object]


class CommandConnect:
    """Command to create Connect feature."""

    def GetResources(self):
        return {
            "Pixmap": getIconPath("Part_JoinConnect.svg"),
            "MenuText": QtCore.QT_TRANSLATE_NOOP("Part_JoinConnect", "Connect Shapes"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "Part_JoinConnect", "Fuses shapes, taking care to preserve voids"
            ),
        }

    def Activated(self):
        if len(FreeCADGui.Selection.getSelectionEx()) >= 1:
            cmdCreateJoinFeature(name="Connect", mode="Connect")
        else:
            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(
                translate(
                    "Part_JoinFeatures",
                    "Select at least two objects, or one or more compounds",
                    None,
                )
            )
            mb.setWindowTitle(translate("Part_JoinFeatures", "Bad Selection", None))
            mb.exec_()

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


# -------------------------- /Connect -----------------------------------------


# -------------------------- Embed --------------------------------------------


def makeEmbed(name):
    """makeEmbed(name): makes an Embed object."""
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", name)
    FeatureEmbed(obj)
    if FreeCAD.GuiUp:
        ViewProviderEmbed(obj.ViewObject)
    return obj


class FeatureEmbed:
    """The Part Embed object."""

    def __init__(self, obj):
        obj.addProperty("App::PropertyLink", "Base", "Embed", "Object to embed into.", locked=True)
        obj.addProperty("App::PropertyLink", "Tool", "Embed", "Object to be embedded.", locked=True)
        obj.addProperty(
            "App::PropertyBool",
            "Refine",
            "Embed",
            "True = refine resulting shape. False = output as is.",
            locked=True,
        )
        obj.Refine = getParamRefine()
        obj.addProperty(
            "App::PropertyLength",
            "Tolerance",
            "Embed",
            "Tolerance when intersecting (fuzzy value). "
            "In addition to tolerances of the shapes.",
            locked=True,
        )

        obj.Proxy = self
        self.Type = "FeatureEmbed"

    def execute(self, selfobj):
        rst = JoinAPI.embed_legacy(selfobj.Base.Shape, selfobj.Tool.Shape, selfobj.Tolerance)
        if selfobj.Refine:
            rst = rst.removeSplitter()
        selfobj.Shape = rst


class ViewProviderEmbed:
    """A View Provider for the Part Embed feature."""

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/booleans/Part_JoinEmbed.svg"

    def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def claimChildren(self):
        return [self.Object.Base, self.Object.Tool]

    def onDelete(self, feature, subelements):
        try:
            self.Object.Base.ViewObject.show()
            self.Object.Tool.ViewObject.show()
        except Exception as err:
            FreeCAD.Console.PrintError("Error in onDelete: " + str(err))
        return True


class CommandEmbed:
    """Command to create Part Embed feature."""

    def GetResources(self):
        return {
            "Pixmap": getIconPath("Part_JoinEmbed.svg"),
            "MenuText": QtCore.QT_TRANSLATE_NOOP("Part_JoinEmbed", "Embed Shapes"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "Part_JoinEmbed", "Fuses one shape into another, taking care to preserve voids"
            ),
        }

    def Activated(self):
        if len(FreeCADGui.Selection.getSelectionEx()) == 2:
            cmdCreateJoinFeature(name="Embed", mode="Embed")
        else:
            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(
                translate(
                    "Part_JoinFeatures",
                    "Select base object, then the object to embed, and then invoke this tool.",
                    None,
                )
            )
            mb.setWindowTitle(translate("Part_JoinFeatures", "Bad Selection", None))
            mb.exec_()

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


# -------------------------- /Embed -------------------------------------------


# -------------------------- Cutout -------------------------------------------


def makeCutout(name):
    """makeCutout(name): makes an Cutout object."""
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", name)
    FeatureCutout(obj)
    if FreeCAD.GuiUp:
        ViewProviderCutout(obj.ViewObject)
    return obj


class FeatureCutout:
    """The Part Cutout object."""

    def __init__(self, obj):
        obj.addProperty("App::PropertyLink", "Base", "Cutout", "Object to be cut.", locked=True)
        obj.addProperty(
            "App::PropertyLink", "Tool", "Cutout", "Object to make cutout for.", locked=True
        )
        obj.addProperty(
            "App::PropertyBool",
            "Refine",
            "Cutout",
            "True = refine resulting shape. False = output as is.",
            locked=True,
        )
        obj.Refine = getParamRefine()
        obj.addProperty(
            "App::PropertyLength",
            "Tolerance",
            "Cutout",
            "Tolerance when intersecting (fuzzy value). In addition to tolerances of the shapes.",
            locked=True,
        )

        obj.Proxy = self
        self.Type = "FeatureCutout"

    def execute(self, selfobj):
        rst = JoinAPI.cutout_legacy(selfobj.Base.Shape, selfobj.Tool.Shape, selfobj.Tolerance)
        if selfobj.Refine:
            rst = rst.removeSplitter()
        selfobj.Shape = rst


class ViewProviderCutout:
    """A View Provider for the Part Cutout feature."""

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/booleans/Part_JoinCutout.svg"

    def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def claimChildren(self):
        return [self.Object.Base, self.Object.Tool]

    def onDelete(self, feature, subelements):
        try:
            self.Object.Base.ViewObject.show()
            self.Object.Tool.ViewObject.show()
        except Exception as err:
            FreeCAD.Console.PrintError("Error in onDelete: " + str(err))
        return True


class CommandCutout:
    """Command to create PartJoinFeature in Cutout mode."""

    def GetResources(self):
        return {
            "Pixmap": getIconPath("Part_JoinCutout.svg"),
            "MenuText": QtCore.QT_TRANSLATE_NOOP("Part_JoinCutout", "Cutout Shape"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "Part_JoinCutout", "Creates a cutout in the selected shape to fit another shape"
            ),
        }

    def Activated(self):
        if len(FreeCADGui.Selection.getSelectionEx()) == 2:
            cmdCreateJoinFeature(name="Cutout", mode="Cutout")
        else:
            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(
                translate(
                    "Part_JoinFeatures",
                    "Select the object to make a cutout in, then the object that should fit into the cutout, and then invoke this tool.",
                    None,
                )
            )
            mb.setWindowTitle(translate("Part_JoinFeatures", "Bad Selection", None))
            mb.exec_()

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


# -------------------------- /Cutout ------------------------------------------


def addCommands():
    FreeCADGui.addCommand("Part_JoinCutout", CommandCutout())
    FreeCADGui.addCommand("Part_JoinEmbed", CommandEmbed())
    FreeCADGui.addCommand("Part_JoinConnect", CommandConnect())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 FreeCAD Team                                       *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENSE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Unit tests for the Property Editor Widgets."""

import unittest
import FreeCAD
from Path.Tool.docobject import DetachedDocumentObject
from Path.Tool.docobject.ui.property import (
    BasePropertyEditorWidget,
    QuantityPropertyEditorWidget,
    BoolPropertyEditorWidget,
    IntPropertyEditorWidget,
    EnumPropertyEditorWidget,
    LabelPropertyEditorWidget,
)


class TestPropertyEditorFactory(unittest.TestCase):
    """Tests the BasePropertyEditorWidget.for_property factory method."""

    def setUp(self):
        # Use the real DetachedDocumentObject
        self.obj = DetachedDocumentObject()
        # Add properties using the DetachedDocumentObject API with correct signature
        self.obj.addProperty("App::PropertyLength", "Length", "Base", "Length property")
        self.obj.Length = FreeCAD.Units.Quantity(10.0)  # Set value separately

        self.obj.addProperty("App::PropertyBool", "IsEnabled", "Base", "Boolean property")
        self.obj.IsEnabled = True  # Set value separately

        self.obj.addProperty("App::PropertyInt", "Count", "Base", "Integer property")
        self.obj.Count = 5  # Set value separately

        self.obj.addProperty("App::PropertyEnumeration", "Mode", "Base", "Enumeration property")
        # Set enums and initial value separately
        self.obj.Mode = ["Auto", "Manual"]

        self.obj.addProperty("App::PropertyString", "Comment", "Base", "String property")
        self.obj.Comment = "Test"  # Set value separately

    def test_quantity_creation(self):
        widget = BasePropertyEditorWidget.for_property(self.obj, "Length")
        self.assertIsInstance(widget, QuantityPropertyEditorWidget)

    def test_bool_creation(self):
        widget = BasePropertyEditorWidget.for_property(self.obj, "IsEnabled")
        self.assertIsInstance(widget, BoolPropertyEditorWidget)

    def test_int_creation(self):
        widget = BasePropertyEditorWidget.for_property(self.obj, "Count")
        self.assertIsInstance(widget, IntPropertyEditorWidget)

    def test_enum_creation(self):
        widget = BasePropertyEditorWidget.for_property(self.obj, "Mode")
        self.assertIsInstance(widget, EnumPropertyEditorWidget)

    def test_label_creation_for_string(self):
        widget = BasePropertyEditorWidget.for_property(self.obj, "Comment")
        self.assertIsInstance(widget, LabelPropertyEditorWidget)

    def test_label_creation_for_invalid_prop(self):
        widget = BasePropertyEditorWidget.for_property(self.obj, "NonExistent")
        self.assertIsInstance(widget, LabelPropertyEditorWidget)


class TestQuantityPropertyEditorWidget(unittest.TestCase):
    """Tests for QuantityPropertyEditorWidget."""

    def setUp(self):
        self.obj = DetachedDocumentObject()
        self.obj.addProperty("App::PropertyLength", "Length", "Base", "Length property")
        self.obj.Length = FreeCAD.Units.Quantity("10.0 mm")
        self.widget = QuantityPropertyEditorWidget(self.obj, "Length")
        # Access the real editor widget
        self.editor = self.widget._editor_widget
        self.widget.updateWidget()

    def test_update_property(self):
        # Check if the real widget's value is updated
        self.assertEqual(self.editor.property("rawValue"), 10.0)
        # Check if the real widget's value and unit are updated
        self.assertEqual(
            self.editor.property("value"), FreeCAD.Units.Quantity(10.0, FreeCAD.Units.Length)
        )

        # Simulate changing the raw value and check if the object's value updates
        self.editor.lineEdit().setText("12.0")
        self.widget.updateProperty()
        self.assertEqual(self.obj.Length.Value, 12.0)
        self.assertEqual(self.obj.Length, FreeCAD.Units.Quantity(12.0, FreeCAD.Units.Length))

        # Try assignment with unit.
        self.editor.lineEdit().setText("15.5 in")
        self.widget.updateProperty()
        updated_value = self.obj.getPropertyByName("Length")
        self.assertIsInstance(updated_value, FreeCAD.Units.Quantity)
        self.assertEqual(updated_value, FreeCAD.Units.Quantity("15.5 in"))


class TestBoolPropertyEditorWidget(unittest.TestCase):
    """Tests for BoolPropertyEditorWidget."""

    def setUp(self):
        self.obj = DetachedDocumentObject()
        self.obj.addProperty("App::PropertyBool", "IsEnabled", "Base", "Boolean property")
        self.obj.IsEnabled = True
        self.widget = BoolPropertyEditorWidget(self.obj, "IsEnabled")
        # Access the real editor widget
        self.editor = self.widget._editor_widget

    def test_update_widget(self):
        self.widget.updateWidget()
        # Check if the real widget's value is updated
        self.assertEqual(self.editor.currentIndex(), 1)  # True is index 1

        self.obj.setPropertyByName("IsEnabled", False)
        self.widget.updateWidget()
        self.assertEqual(self.editor.currentIndex(), 0)  # False is index 0

    def test_update_property(self):
        # Simulate user changing value in the combobox
        self.editor.setCurrentIndex(0)  # Select False
        self.widget._on_index_changed(0)
        self.assertEqual(self.obj.getPropertyByName("IsEnabled"), False)

        self.editor.setCurrentIndex(1)  # Select True
        self.widget._on_index_changed(1)
        self.assertEqual(self.obj.getPropertyByName("IsEnabled"), True)


class TestIntPropertyEditorWidget(unittest.TestCase):
    """Tests for IntPropertyEditorWidget."""

    def setUp(self):
        self.obj = DetachedDocumentObject()
        self.obj.addProperty("App::PropertyInt", "Count", "Base", "Integer property")
        self.obj.Count = 5
        self.widget = IntPropertyEditorWidget(self.obj, "Count")
        # Access the real editor widget
        self.editor = self.widget._editor_widget

    def test_update_widget(self):
        self.widget.updateWidget()
        # Check if the real widget's value is updated
        self.assertEqual(self.editor.value(), 5)

        self.obj.setPropertyByName("Count", 100)
        self.widget.updateWidget()
        self.assertEqual(self.editor.value(), 100)

    def test_update_property(self):
        # Simulate user changing value in the spinbox
        self.editor.setValue(42)
        self.widget.updateProperty()
        self.assertEqual(self.obj.getPropertyByName("Count"), 42)


class TestEnumPropertyEditorWidget(unittest.TestCase):
    """Tests for EnumPropertyEditorWidget."""

    def setUp(self):
        self.obj = DetachedDocumentObject()
        self.obj.addProperty("App::PropertyEnumeration", "Mode", "Base", "Enumeration property")
        self.obj.Mode = ["Auto", "Manual", "Semi"]  # Set enums and initial value
        self.widget = EnumPropertyEditorWidget(self.obj, "Mode")
        # Access the real editor widget
        self.editor = self.widget._editor_widget

    def test_populate_enum(self):
        # Check if the real widget is populated
        self.assertEqual(self.editor.count(), 3)
        self.assertEqual(self.editor.itemText(0), "Auto")
        self.assertEqual(self.editor.itemText(1), "Manual")
        self.assertEqual(self.editor.itemText(2), "Semi")

    def test_update_widget(self):
        self.widget.updateWidget()
        # Check if the real widget's value is updated
        self.assertEqual(self.editor.currentText(), "Auto")

        self.obj.setPropertyByName("Mode", "Manual")
        self.widget.updateWidget()
        self.assertEqual(self.editor.currentText(), "Manual")

    def test_update_property(self):
        # Simulate user changing value in the combobox
        self.editor.setCurrentIndex(1)  # Select Manual
        self.widget._on_index_changed(1)
        self.assertEqual(self.obj.getPropertyByName("Mode"), "Manual")

        self.editor.setCurrentIndex(2)  # Select Semi
        self.widget._on_index_changed(2)
        self.assertEqual(self.obj.getPropertyByName("Mode"), "Semi")


class TestLabelPropertyEditorWidget(unittest.TestCase):
    """Tests for LabelPropertyEditorWidget."""

    def setUp(self):
        self.obj = DetachedDocumentObject()
        self.obj.addProperty("App::PropertyString", "Comment", "Base", "String property")
        self.obj.Comment = "Test Comment"
        self.widget = LabelPropertyEditorWidget(self.obj, "Comment")
        # Access the real editor widget
        self.editor = self.widget._editor_widget

    def test_update_widget(self):
        self.widget.updateWidget()
        # Check if the real widget's value is updated
        self.assertEqual(self.editor.text(), "Test Comment")

        self.obj.setPropertyByName("Comment", "New Comment")
        self.widget.updateWidget()
        self.assertEqual(self.editor.text(), "New Comment")

    def test_update_property_is_noop(self):
        # updateProperty should do nothing for a read-only label
        self.widget.updateProperty()
        # No assertions needed, just ensure it doesn't raise errors


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2023 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create Clone objects.

The clone is basically a simple copy of the `Shape` of another object,
whether that is a Draft object or any other 3D object.

The Clone's `Shape` can be scaled in size in any direction.

This implementation was developed before the `App::Link` object was created.
In many cases using `App::Link` makes more sense, as this object is
more memory efficient as it reuses the same internal `Shape`
instead of creating a copy of it.
"""
## @package gui_clone
# \ingroup draftguitools
# \brief Provides GUI tools to create Clone objects.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils
import draftutils.todo as todo
from draftutils.messages import _msg, _wrn
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Clone(gui_base_original.Modifier):
    """Gui Command for the Clone tool."""

    def __init__(self):
        super().__init__()
        self.moveAfterCloning = False

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Clone",
            "Accel": "C, L",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Clone", "Clone"),
            "ToolTip": QT_TRANSLATE_NOOP("Draft_Clone", "Creates a clone of the selected objects"),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated(name="Clone")
        if not self.ui:
            return
        if not Gui.Selection.getSelection():
            self.ui.selectUi(on_close_call=self.finish)
            _msg(translate("draft", "Select an object to clone"))
            self.call = self.view.addEventCallback("SoEvent", gui_tool_utils.selectObject)
        else:
            self.proceed()

    def proceed(self):
        """Proceed with the command if objects were selected."""
        objs = Gui.Selection.getSelection()
        if not objs:
            self.finish()
            return
        objs_shape = [obj for obj in objs if hasattr(obj, "Shape")]
        if not objs_shape:
            _wrn(translate("draft", "Cannot clone objects without a shape, aborting"))
            self.finish()
            return
        elif len(objs_shape) < len(objs):
            _wrn(translate("draft", "Cannot clone objects without a shape, skipping them"))

        Gui.addModule("Draft")
        App.ActiveDocument.openTransaction(translate("Draft", "Create Clone"))
        for idx, obj in enumerate(objs_shape):
            cmd = "Draft.make_clone(FreeCAD.ActiveDocument." + obj.Name + ")"
            Gui.doCommand("clone" + str(idx) + " = " + cmd)
        App.ActiveDocument.commitTransaction()
        App.ActiveDocument.recompute()

        Gui.Selection.clearSelection()
        objs = App.ActiveDocument.Objects
        for i in range(len(objs_shape)):
            Gui.Selection.addSelection(objs[-(1 + i)])
        self.finish()

    def finish(self):
        """Terminate the operation of the tool."""
        if self.call is not None:
            self.end_callbacks(self.call)
        super().finish()
        if self.moveAfterCloning:
            todo.ToDo.delay(Gui.runCommand, "Draft_Move")


Draft_Clone = Clone
Gui.addCommand("Draft_Clone", Clone())

## @}
